# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
# %matplotlib widget

# %% [markdown]
# # Applications

# %% [markdown] tags=[]
# ## Libraries

# %% [markdown]
# Import basic libraries and `kdfeatures`.

# %%
import numpy as np
import matplotlib.pyplot as plt

import kdfeatures.applications.mlb as mlb
import kdfeatures.applications.nba as nba
import kdfeatures.applications.nfl as nfl
import kdfeatures.bandwidth as kbw
import kdfeatures.density as kde
import kdfeatures.plot as kpl

# %% [markdown]
# ## Configuration

# %% [markdown]
# Plotting.

# %%
plt.rcParams.update({"figure.max_open_warning": 0})


def save_fig(file_prefix: str, file_suffix: str):
    plt.savefig(f"jpg/{file_prefix}_{file_suffix}.jpg", bbox_inches="tight")


# %% [markdown]
# Random number generation.

# %%
rand = np.random.RandomState(seed=1)

# %% [markdown]
# ## Univariate densities: NFL

# %%
def fit_kde_passing_yds(passing_yds: np.ndarray) -> kde.KDE:
    h = kbw.univariate_smoothed_cv_bandwidth(
        sample=passing_yds,
        r=2,
        num_stages=3,
        print_progress=True,
        print_every_num_iter=10,
    )

    f = kde.univariate_kde(sample=passing_yds, h=h)

    return f


def plot_passing_yds(passing_yds_kde: kde.KDE, file_prefix: str):
    file_prefix = "nfl_yds_" + file_prefix

    nfl.plot_scatter_yards(
        kde=passing_yds_kde,
        contour_feature=kpl.PDFLaplacian(),
        scatter_feature=kpl.PDFUnsignedHeight(),
        rand=rand,
    )

    save_fig(file_prefix, "scatter")

    kpl.plot_surface_1d(
        pdf=passing_yds_kde,
        xmin=-10,
        xmax=150,
        xnum=1000,
        area_feature=kpl.PDFLaplacian(),
        fsolve_x0_size=100,
    )

    save_fig(file_prefix, "surface")


# %%
kde_brady_2014 = fit_kde_passing_yds(nfl.passing_yards_by_tom_brady_2014())

# %%
kde_brady_2016 = fit_kde_passing_yds(nfl.passing_yards_by_tom_brady_2016())

# %%
kde_brady_2018 = fit_kde_passing_yds(nfl.passing_yards_by_tom_brady_2018())

# %%
plot_passing_yds(kde_brady_2014, "brady_2014")

# %%
plot_passing_yds(kde_brady_2016, "brady_2016")

# %%
plot_passing_yds(kde_brady_2018, "brady_2018")

# %% [markdown]
# ## Bivariate densities: NBA

# %%
def fit_kde_shots(shots: np.ndarray) -> kde.KDE:
    h = kbw.smoothed_cv_bandwidth(
        sample=shots,
        r=2,
        bandwidth_type=kbw.Bandwidths.scalar,
        num_stages=1,
        print_progress=True,
        print_loss=True,
        print_every_num_iter=10,
        maxiter=50,
    )

    f = kde.KDE(
        sample=shots,
        bandwidth=h,
    )

    return f


def plot_scatter_shots(
    kde_shots: kde.KDE, file_prefix: str, num_points: int = 201
):
    file_prefix = "nba_shots_" + file_prefix

    def plot_scatter(feature: kpl.PDFSignedFeature, color: str):
        nba.plot_scatter_shots(
            kde=kde_shots,
            contour_features=[feature],
            colors=[color],
            linestyles=["solid"],
            numx=num_points,
            numy=num_points,
            fill_contour=True,
            contour_opacity=0.5,
        )

    plot_scatter(
        feature=kpl.PDFSignedHeight(kde_shots.quantile(0.35)), color="r"
    )

    save_fig(file_prefix, "scatter_hdr35")

    plot_scatter(
        feature=kpl.PDFSignedHeight(kde_shots.quantile(0.95)),
        color="darkorange",
    )

    save_fig(file_prefix, "scatter_hdr95")

    plot_scatter(feature=kpl.PDFLargestHessianEigenvalue(), color="m")

    save_fig(file_prefix, "scatter_concave")

    plot_scatter(feature=kpl.PDFLaplacian(), color="b")

    save_fig(file_prefix, "scatter_laplacian")


def plot_confidence_shots(
    kde_shots: kde.KDE, file_prefix: str, num_bootstrap_samples: int = 100
):
    file_prefix = "nba_shots_" + file_prefix

    def plot_confidence_sets(feature: kpl.PDFSignedFeature, confidence: float):
        nba.plot_confidence_shots(
            kde=kde_shots,
            confidence=confidence,
            num_bootstrap_samples=num_bootstrap_samples,
            feature=feature,
            rand=rand,
        )

    plot_confidence_sets(feature=kpl.PDFLaplacian(), confidence=0.9)
    save_fig(file_prefix, "confidence_laplacian")

    plot_confidence_sets(
        feature=kpl.PDFLargestHessianEigenvalue(), confidence=0.9
    )

    save_fig(file_prefix, "confidence_concave")


def plot_scatter_surface_shots(
    kde_shots: kde.KDE, file_prefix: str, num_points: int = 201
):
    file_prefix = "nba_shots_" + file_prefix

    nba.plot_scatter_shots(
        kde=kde_shots,
        contour_features=[
            kpl.PDFLaplacian(),
            kpl.PDFLargestHessianEigenvalue(),
        ],
        scatter_feature=kpl.PDFUnsignedHeight(),
        colors=["b", "m"],
        linestyles=["solid", "solid"],
        numx=num_points,
        numy=num_points,
    )

    save_fig(file_prefix, "scatter")

    nba.plot_surface_shots(
        kde=kde_shots,
        numx=num_points,
        numy=num_points,
        area_feature=kpl.PDFLargestHessianEigenvalue(),
    )

    save_fig(file_prefix, "surface_concavity")

    nba.plot_surface_shots(
        kde=kde_shots,
        numx=num_points,
        numy=num_points,
        area_feature=kpl.PDFLaplacian(),
    )

    save_fig(file_prefix, "surface_laplacian")


# %%
kde_curry = fit_kde_shots(nba.stephen_curry_shots())

# %%
kde_durant = fit_kde_shots(nba.kevin_durant_shots())

# %%
kde_harden = fit_kde_shots(nba.james_harden_shots())

# %%
plot_scatter_shots(kde_curry, "curry")

# %%
plot_scatter_surface_shots(kde_curry, "curry")

# %%
plot_confidence_shots(kde_curry, "curry")

# %%
plot_scatter_surface_shots(kde_durant, "durant")

# %%
plot_confidence_shots(kde_durant, "durant")

# %%
plot_scatter_surface_shots(kde_harden, "harden")

# %%
plot_confidence_shots(kde_harden, "harden")

# %% [markdown] tags=[]
# ## Trivariate densities: MLB

# %% [markdown]
# Auxiliary functions.

# %%
def fit_kde_pitches(pitches: np.ndarray) -> kde.KDE:
    h = kbw.gaussian_mixture_bandwidth(
        pitches,
        r=2,
        num_components=3,
        bandwidth_type=kbw.Bandwidths.scalar,
        rand=rand,
        print_progress=True,
        print_loss=True,
        print_every_num_iter=10,
    )

    f = kde.KDE(
        sample=pitches,
        bandwidth=h,
    )

    return f


# %%
def plot_pitches(
    kde_pitches: kde.KDE,
    min_pitch_speed: float,
    max_pitch_speed: float,
    strike_zone_bottom: float,
    strike_zone_top: float,
    file_prefix: str,
):
    file_prefix = "mlb_pitches_" + file_prefix

    mlb.plot_pitches(
        kde=kde_pitches,
        min_pitch_speed=min_pitch_speed,
        max_pitch_speed=max_pitch_speed,
        strike_zone_bottom=strike_zone_bottom,
        strike_zone_top=strike_zone_top,
        contour_features=[
            kpl.PDFLaplacian(),
            kpl.PDFLargestHessianEigenvalue(),
        ],
        scatter_feature=kpl.PDFUnsignedHeight(),
        contour_color_maps=["Greys", "Blues"],
        contour_transparencies=[0.2, 0.8],
    )

    save_fig(file_prefix, "scatter_surface")


# %%
kershaw_pitches, kershaw_sz_bot, kershaw_sz_top = mlb.clayton_kershaw_pitches()

# %%
kde_kershaw = fit_kde_pitches(kershaw_pitches)

# %%
plot_pitches(
    kde_pitches=kde_kershaw,
    min_pitch_speed=65.0,
    max_pitch_speed=95.0,
    strike_zone_bottom=kershaw_sz_bot,
    strike_zone_top=kershaw_sz_top,
    file_prefix="kershaw",
)

# %%
(
    verlander_pitches,
    verlander_sz_bot,
    verlander_sz_top,
) = mlb.justin_verlander_pitches()

# %%
kde_verlander = fit_kde_pitches(verlander_pitches)

# %%
plot_pitches(
    kde_pitches=kde_verlander,
    min_pitch_speed=70.0,
    max_pitch_speed=105.0,
    strike_zone_bottom=verlander_sz_bot,
    strike_zone_top=verlander_sz_top,
    file_prefix="verlander",
)

# %%
scherzer_pitches, scherzer_sz_bot, scherzer_sz_top = mlb.max_scherzer_pitches()

# %%
kde_scherzer = fit_kde_pitches(scherzer_pitches)

# %%
plot_pitches(
    kde_pitches=kde_scherzer,
    min_pitch_speed=70.0,
    max_pitch_speed=105.0,
    strike_zone_bottom=scherzer_sz_bot,
    strike_zone_top=scherzer_sz_top,
    file_prefix="scherzer",
)
