# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %%
# %matplotlib widget

# %% [markdown] tags=[]
# # Preliminary curvature graphing

# %% [markdown] tags=[]
# ## Libraries

# %% [markdown]
# Import basic libraries and `kdfeatures`.

# %%
import matplotlib.pyplot as plt
import numpy as np

import kdfeatures.bandwidth as kbw
import kdfeatures.density as kde
import kdfeatures.plot as kpl

# %% [markdown] tags=[]
# ## Configuration

# %% [markdown]
# Plotting.

# %%
plt.rcParams.update({"figure.max_open_warning": 0})


def save_fig(file: str):
    plt.savefig(f"jpg/{file}.jpg", bbox_inches="tight")


# %% [markdown]
# Random number generation

# %%
rand = np.random.RandomState(seed=1)


# %% [markdown] tags=[]
# ## Univariate densities

# %% [markdown]
# Auxiliary functions.

# %%
def plot_surface_1d(f: kde.PDF):
    kpl.plot_surface_1d(
        pdf=f,
        xmin=-3,
        xmax=7,
        xnum=1000,
        area_feature=kpl.PDFLargestHessianEigenvalue(),
    )


def plot_sample_1d(f: kde.PDF, sample: np.ndarray = None):
    kpl.plot_sample_1d(
        pdf=f,
        sample=sample,
        xmin=-3,
        xmax=7,
        size=1000,
        rand=rand,
        contour_feature=kpl.PDFLargestHessianEigenvalue(),
        scatter_feature=kpl.PDFUnsignedHeight(),
    )


# %% [markdown] tags=[]
# ### Examples

# %% [markdown]
# + Mixture with two components.

# %%
mixture_1d = kde.Mixture(
    [
        kde.univariate_gaussian(mean=0.0, sigma=1.0),
        kde.univariate_gaussian(mean=4.0, sigma=1.0),
    ],
    [0.7, 0.3],
)

# %%
plot_surface_1d(mixture_1d)

# %%
plot_sample_1d(mixture_1d)

# %% [markdown]
# + KDE of the former.

# %%
sample_mixture_1d = mixture_1d.sample(size=1000, rand=rand)

# %%
bw_mixture_1d = kbw.smoothed_cv_bandwidth(
    sample=sample_mixture_1d,
    r=2,
    num_stages=10,
    print_progress=True,
    print_every_num_iter=5,
)

# %% tags=[]
kde_1d = kde.KDE(
    sample=sample_mixture_1d,
    bandwidth=bw_mixture_1d,
)

# %%
plot_surface_1d(kde_1d)

# %%
plot_sample_1d(kde_1d, sample=sample_mixture_1d)


# %% [markdown] tags=[]
# ## Bivariate densities

# %% [markdown]
# Define auxiliary functions.

# %% [markdown]
# + Gaussian mixture generation.

# %%
def mixture(x: float, corr: float, w: float) -> kde.Mixture:
    return kde.Mixture(
        [
            kde.bivariate_gaussian(mean_x=-x, corr=corr),
            kde.bivariate_gaussian(mean_x=x, corr=-corr),
        ],
        [w, 1 - w],
    )


# %% [markdown]
# + Plotting.

# %%
def plot_surface(
    f: kde.PDF,
    num_points: int,
    feature: kpl.PDFFeature,
    **kwargs,
):
    kpl.plot_surface_2d(
        pdf=f,
        numx=num_points,
        numy=num_points,
        xstride=1,
        ystride=1,
        area_feature=feature,
        **kwargs,
    )


def plot_surfaces(
    f: kde.PDF,
    num_points: int,
    file_eigenvalue: str = None,
    file_laplacian: str = None,
    **kwargs,
):
    plot_surface(
        f=f,
        num_points=num_points,
        feature=kpl.PDFLargestHessianEigenvalue(),
        **kwargs,
    )

    if file_eigenvalue is not None:
        save_fig(file_eigenvalue)

    plot_surface(
        f=f,
        num_points=num_points,
        feature=kpl.PDFMeanCurvature(),
        **kwargs,
    )

    if file_laplacian is not None:
        save_fig(file_laplacian)


def plot_sample(
    f: kde.PDF,
    numx: int,
    numy: int,
    sample: np.ndarray = None,
    size: int = None,
    file: str = None,
    **kwargs,
):
    kpl.plot_sample_2d(
        pdf=f,
        sample=sample,
        size=size,
        scatter_feature=kpl.PDFUnsignedHeight(),
        contour_features=[
            kpl.PDFLargestHessianEigenvalue(),
            kpl.PDFMeanCurvature(),
        ],
        contour_colors=["b", "c"],
        numx=numx,
        numy=numy,
        rand=rand,
        contour_linewidth=3.0,
        **kwargs,
    )

    if file is not None:
        save_fig(file)


# %% [markdown] tags=[]
# ### Examples

# %% [markdown] tags=[]
# #### Mixtures

# %% [markdown]
# 1. Big mountain alongside a small one.

# %%
mixture_2d_1 = mixture(x=2.5, w=0.9, corr=0.0)

# %%
plot_surfaces(
    f=mixture_2d_1, xmin=-5.0, xmax=5.0, ymin=-5.0, ymax=5.0, num_points=201
)

# %%
plot_sample(f=mixture_2d_1, size=1000, numx=101, numy=101)

# %% [markdown]
# 2. Two close big mountains.

# %%
mixture_2d_2 = mixture(x=1.1, w=0.5, corr=0.0)

# %%
plot_surfaces(
    f=mixture_2d_2, xmin=-5.0, xmax=5.0, ymin=-5.0, ymax=5.0, num_points=201
)

# %%
plot_sample(f=mixture_2d_2, size=1000, numx=101, numy=101)

# %% [markdown]
# 3. Big mountain alongside a small one, slight angle.

# %%
mixture_2d_3 = mixture(x=1.1, w=0.9, corr=0.2)

# %%
plot_surfaces(
    f=mixture_2d_3, xmin=-5.0, xmax=5.0, ymin=-5.0, ymax=5.0, num_points=201
)

# %%
plot_sample(f=mixture_2d_3, size=1000, numx=101, numy=101)

# %% [markdown]
# 4. Two close big mountains, pronounced angle.

# %%
mixture_2d_4 = mixture(x=1.5, w=0.5, corr=-0.7)

# %%
plot_surfaces(
    f=mixture_2d_4,
    xmin=-5.0,
    xmax=5.0,
    ymin=-5.0,
    ymax=5.0,
    num_points=201,
    file_eigenvalue="mixture_2d_eigenvalue",
    file_laplacian="mixture_2d_mean_curvature",
)

# %%
plot_sample(
    f=mixture_2d_4,
    size=1000,
    numx=101,
    numy=101,
    xmin=-4.0,
    xmax=4.0,
    ymin=-4.0,
    ymax=4.0,
    file="mixture_2d_sample",
)

# %% [markdown]
# 5. Ray and Lindsay's example.

# %%
mixture_ray_lindsay = kde.Mixture(
    [
        kde.bivariate_gaussian(sigma_y=np.sqrt(0.05)),
        kde.bivariate_gaussian(mean_x=1.0, mean_y=1.0, sigma_x=np.sqrt(0.05)),
    ],
    [0.5, 0.5],
)

# %%
plot_surfaces(
    f=mixture_ray_lindsay,
    xmin=-5.0,
    xmax=5.0,
    ymin=-4.0,
    ymax=6.0,
    num_points=501,
)

# %%
plot_sample(
    f=mixture_ray_lindsay,
    size=10_000,
    xmin=-5.0,
    xmax=5.0,
    ymin=-4.0,
    ymax=6.0,
    numx=501,
    numy=501,
    scatter_size=10.0,
)

# %% [markdown] tags=[]
# #### KDE

# %% [markdown]
# KDE of the former mixture.

# %%
sample_mixture_2d_4 = mixture_2d_4.sample(size=1000, rand=rand)

# %%
bw_mixture_2d_4 = kbw.smoothed_cv_bandwidth(
    sample=sample_mixture_2d_4,
    r=2,
    num_stages=3,
    print_progress=True,
    print_every_num_iter=10,
    maxiter=50,
)

# %%
kde_2d = kde.KDE(sample=sample_mixture_2d_4, bandwidth=bw_mixture_2d_4)

# %%
plot_surfaces(
    f=kde_2d, xmin=-5.0, xmax=5.0, ymin=-5.0, ymax=5.0, num_points=101
)

# %%
plot_sample(f=kde_2d, numx=101, numy=101, sample=sample_mixture_2d_4)


# %% [markdown] tags=[]
# ## Trivariate densities

# %% [markdown]
# Auxiliary functions.

# %%
def plot_sample_3d(f: kde.PDF, num: int = 101, sample: np.ndarray = None):
    kpl.plot_sample_3d(
        pdf=f,
        contour_features=[
            kpl.PDFLaplacian(),
            kpl.PDFLargestHessianEigenvalue(),
        ],
        scatter_feature=kpl.PDFUnsignedHeight(),
        contour_color_maps=["Greys", "Blues"],
        contour_transparencies=[0.2, 0.8],
        xmin=-5.0,
        xmax=5.0,
        ymin=-5.0,
        ymax=5.0,
        zmin=-5.0,
        zmax=5.0,
        numx=num,
        numy=num,
        numz=num,
        sample=sample,
        size=1000,
        rand=rand,
    )


# %% [markdown]
# ### Examples

# %% [markdown]
# + Mixture with two components.

# %%
mixture_3d = kde.Mixture(
    pdfs=[
        kde.trivariate_gaussian(mean_x=-1.5, corr_xy=-0.5),
        kde.trivariate_gaussian(mean_x=1.5, corr_xy=0.5),
    ],
    weights=[1 / 2, 1 / 2],
)

# %%
plot_sample_3d(mixture_3d)

# %% [markdown]
# + KDE of the former.

# %%
sample_mixture_3d = mixture_3d.sample(size=1000, rand=rand)

# %%
bw_mixture_3d = kbw.smoothed_cv_bandwidth(
    sample=sample_mixture_3d,
    r=2,
    num_stages=2,
    print_progress=True,
    print_every_num_iter=10,
    maxiter=50,
)

# %%
kde_3d = kde.KDE(sample=sample_mixture_3d, bandwidth=bw_mixture_3d)

# %%
plot_sample_3d(kde_3d, num=21, sample=sample_mixture_3d)
