"""
Downloads pitching data from an MLB player.
"""
import argparse
import sys
from typing import Final, List

import pandas as pd
from pybaseball import playerid_lookup, statcast_pitcher

MLB_GAME_ID: Final = "game_pk"
MLB_AT_BAT_NUMBER: Final = "at_bat_number"
MLB_PITCH_NUMBER: Final = "pitch_number"
MLB_PLAYER_NAME: Final = "player_name"
MLB_BATTER_ID: Final = "batter"
MLB_INNING: Final = "inning"
MLB_PITCH_NAME: Final = "pitch_name"
MLB_PLATE_X: Final = "plate_x"
MLB_PLATE_Z: Final = "plate_z"
MLB_RELEASE_SPEED: Final = "release_speed"
MLB_RELEASE_SPIN_RATE: Final = "release_spin_rate"
MLB_SPIN_AXIS: Final = "spin_axis"
MLB_PITCHER_HAND: Final = "p_throws"
MLB_STRIKE_ZONE_BOTTOM: Final = "sz_bot"
MLB_STRIKE_ZONE_TOP: Final = "sz_top"
MLB_PITCH_ZONE: Final = "zone"
MLB_PITCH_RESULT: Final = "type"
MLB_PITCH_EVENTS: Final = "events"


def find_mlb_player(
    player_first_name: str,
    player_last_name: str,
    raise_error_not_unique: bool = True,
) -> List[int]:
    """
    Finds the player identifiers associated to a name.

    :param player_first_name: Player's first name.
    :param player_last_name: Player's last name.
    :param raise_error_not_unique: Whether to raise an exception if more than
     one identifier is found.
    :return: A list of player identifiers corresponding to the input names.
    """
    player: Final = playerid_lookup(
        last=player_last_name, first=player_first_name
    )

    if raise_error_not_unique and len(player.index) > 1:
        raise ValueError("More than one player found")

    return [int(player_id) for player_id in player.key_mlbam]


def download_mlb_pitches(
    start_date: str,
    end_date: str,
    player_first_name: str = None,
    player_last_name: str = None,
    player_id: int = None,
) -> pd.DataFrame:
    """
    Retrieves pitching data from MLB's API by a player's identifier or names.

    :param start_date: First pitching date.
    :param end_date: Last pitching date.
    :param player_first_name: Player's first name.
    :param player_last_name: Player's last name.
    :param player_id: Player's unique identifier.
    :return: A data frame with pitching data.
    """

    if player_id is None:
        if player_first_name is None or player_last_name is None:
            raise ValueError(
                "Either a player id. or a player name must be provided"
            )

        player_id = find_mlb_player(
            player_first_name=player_first_name,
            player_last_name=player_last_name,
            raise_error_not_unique=True,
        )[0]

    pitches: Final = statcast_pitcher(
        start_dt=start_date,
        end_dt=end_date,
        player_id=player_id,
    )

    return pitches[
        [
            MLB_GAME_ID,
            MLB_AT_BAT_NUMBER,
            MLB_PITCH_NUMBER,
            MLB_PLAYER_NAME,
            MLB_BATTER_ID,
            MLB_INNING,
            MLB_PITCH_NAME,
            MLB_PLATE_X,
            MLB_PLATE_Z,
            MLB_RELEASE_SPEED,
            MLB_RELEASE_SPIN_RATE,
            MLB_SPIN_AXIS,
            MLB_PITCHER_HAND,
            MLB_STRIKE_ZONE_BOTTOM,
            MLB_STRIKE_ZONE_TOP,
            MLB_PITCH_ZONE,
            MLB_PITCH_RESULT,
            MLB_PITCH_EVENTS,
        ]
    ]


def save_mlb_pitches(
    start_date: str,
    end_date: str,
    csv: str,
    player_first_name: str = None,
    player_last_name: str = None,
    player_id: int = None,
):
    """
    Downloads MLB pitching data and saves it in a CSV file.

    :param start_date: First pitching date.
    :param end_date: Last pitching date.
    :param csv: Name of the CSV where data should be saved.
    :param player_first_name: Player's first name.
    :param player_last_name: Player's last name.
    :param player_id: Player's unique identifier.
    """
    df = download_mlb_pitches(
        start_date=start_date,
        end_date=end_date,
        player_first_name=player_first_name,
        player_last_name=player_last_name,
        player_id=player_id,
    )

    df.to_csv(csv, index=False)


def _config_args_parser() -> argparse.ArgumentParser:
    """
    Builds and returns the arguments parser for this program.

    :return: The arguments parser for this program.
    """
    parser = argparse.ArgumentParser(
        description="Downloads pitching data from an MLB player into a CSV file"
    )

    parser.add_argument(
        "--player-first-name",
        type=str,
        help="Player first name",
    )

    parser.add_argument(
        "--player-last-name",
        type=str,
        help="Player last name",
    )

    parser.add_argument(
        "--player-id",
        type=str,
        help="Player identifier",
    )

    parser.add_argument(
        "--start-date",
        type=str,
        required=True,
        help="Pitching start date",
    )

    parser.add_argument(
        "--end-date",
        type=str,
        required=True,
        help="Pitching end date",
    )

    parser.add_argument(
        "--csv",
        type=str,
        required=True,
        help="CSV file",
    )

    return parser


def main():
    """
    Application entry point.
    """

    # Parse args.

    parser = _config_args_parser()
    args = parser.parse_args(sys.argv[1:])

    # Download CSV.

    save_mlb_pitches(
        player_first_name=args.player_first_name,
        player_last_name=args.player_last_name,
        player_id=args.player_id,
        start_date=args.start_date,
        end_date=args.end_date,
        csv=args.csv,
    )


if __name__ == "__main__":
    main()
