"""
MLB data and experiments.
"""
from typing import Final, List

import numpy as np
from mpl_toolkits.mplot3d.art3d import Poly3DCollection

from kdfeatures.density import KDE
from kdfeatures.plot import (
    PDFFeature,
    PDFSignedFeature,
    plot_sample_3d,
)


def plot_pitches(
    kde: KDE,
    min_pitch_speed: float = 65.0,
    max_pitch_speed: float = 105.0,
    strike_zone_bottom: float = None,
    strike_zone_top: float = None,
    scatter_feature: PDFFeature = None,
    scatter_size: int = 1,
    contour_features: List[PDFSignedFeature] = None,
    contour_color_maps: List[str] = None,
    contour_transparencies: List[float] = None,
    num_points: int = 21,
):
    """
    Plots scattered pitches.

    :param kde: Kernel density estimator for the pitches.
    :param min_pitch_speed: Minimum pitching speed.
    :param max_pitch_speed: Maximum pitching speed.
    :param strike_zone_bottom: Bottom z coordinate of the strike zone.
    :param strike_zone_top: Top z coordinate of the strike zone.
    :param scatter_feature: Feature to color scatter dots.
    :param scatter_size: Size of pitch markers.
    :param contour_features: Features to determine contours.
    :param contour_color_maps: Color maps for contours.
    :param contour_transparencies: Transparency levels of contours.
    :param num_points: Number of points in each axis.
    """
    # Scatter plot.

    ax = plot_sample_3d(
        pdf=kde,
        sample=kde.underlying_sample,
        scatter_feature=scatter_feature,
        scatter_size=scatter_size,
        contour_features=contour_features,
        contour_color_maps=contour_color_maps,
        contour_transparencies=contour_transparencies,
        ymin=min_pitch_speed,
        ymax=max_pitch_speed,
        numx=num_points,
        numy=num_points,
        numz=num_points,
        elevation=10.0,
        azimuth=30.0,
    )

    xlim_left: Final = -3.0
    xlim_right: Final = 3.0
    zlim_left: Final = -1.0
    zlim_right: Final = 6.0

    ax.set_xticklabels([])
    ax.set_zticklabels([])

    ax.set_xlim(xlim_left, xlim_right)
    ax.set_zlim(zlim_left, zlim_right)
    ax.set_ylim(min_pitch_speed, max_pitch_speed)

    ax.set_xticks(list(np.arange(xlim_left, xlim_right + 1, 1)))
    ax.set_zticks(list(np.arange(zlim_left, zlim_right + 1, 1)))
    ax.set_yticks(list(np.arange(min_pitch_speed, max_pitch_speed + 5, 5)))

    ax.set_ylabel("mph")

    # Strike zone polygon.

    if strike_zone_bottom is not None and strike_zone_top is not None:
        x_max: Final = 0.83
        x_min: Final = -x_max
        y_max = max_pitch_speed

        strike_zone = Poly3DCollection(
            [
                [
                    (x_min, y_max, strike_zone_bottom),
                    (x_max, y_max, strike_zone_bottom),
                    (x_max, y_max, strike_zone_top),
                    (x_min, y_max, strike_zone_top),
                ]
            ],
        )

        strike_zone.set_facecolor("w")
        strike_zone.set_edgecolor("k")
        strike_zone.set_linewidth(2.0)
        strike_zone.set_alpha(0.5)

        ax.add_collection3d(strike_zone)
