"""
Access to downloaded MLB pitching data.
"""
import os
from typing import Final, Tuple

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from kdfeatures.applications.mlb.api import (
    MLB_PLATE_X,
    MLB_PLATE_Z,
    MLB_RELEASE_SPEED,
    MLB_STRIKE_ZONE_BOTTOM,
    MLB_STRIKE_ZONE_TOP,
)
from kdfeatures.util.matrix import remove_nan_rows

_CLAYTON_KERSHAW_PITCHES_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "clayton_kershaw_2019.csv")
)

_JUSTIN_VERLANDER_PITCHES_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "justin_verlander_2019.csv")
)

_MAX_SCHERZER_PITCHES_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "max_scherzer_2019.csv")
)


def read_pitches(csv: str) -> Tuple[np.ndarray, float, float]:
    """
    Reads pitching data from a CSV file. The first and third pitching columns
    correspond to x and z coordinates seen from home plate, while the second
    column is the release speed.

    :param csv: CSV file to read data from.
    :return: A 3-tuple, where the first component is the pitching information
     and the second and third correspond to average strike zone top and
     bottom positions, respectively.
    """
    column_names: Final = [
        MLB_PLATE_X,
        MLB_RELEASE_SPEED,
        MLB_PLATE_Z,
        MLB_STRIKE_ZONE_TOP,
        MLB_STRIKE_ZONE_BOTTOM,
    ]

    df: Final = pd.read_csv(
        csv,
        usecols=column_names,
    )

    columns: Final = [np.array(df[name]).astype(float) for name in column_names]

    sample: Final = remove_nan_rows(np.column_stack(columns))

    return (
        sample[:, 0:3],
        float(np.mean(sample[:, 3])),
        float(np.mean(sample[:, 4])),
    )


def clayton_kershaw_pitches() -> Tuple[np.ndarray, float, float]:
    """
    Returns pitching data by Clayton Kershaw in the 2019 season.

    :return: Pitching data by Clayton Kershaw in the 2019 season.
    """
    return read_pitches(csv=_CLAYTON_KERSHAW_PITCHES_CSV_PATH)


def justin_verlander_pitches() -> Tuple[np.ndarray, float, float]:
    """
    Returns pitching data by Justin Verlander in the 2019 season.

    :return: Pitching data by Justin Verlander in the 2019 season.
    """
    return read_pitches(csv=_JUSTIN_VERLANDER_PITCHES_CSV_PATH)


def max_scherzer_pitches() -> Tuple[np.ndarray, float, float]:
    """
    Returns pitching data by Max Scherzer in the 2019 season.

    :return: Pitching data by Max Scherzer in the 2019 season.
    """
    return read_pitches(csv=_MAX_SCHERZER_PITCHES_CSV_PATH)
