"""
Downloads shot data from an NBA player.
"""
import argparse
import json
import sys
from typing import Final, List

import pandas as pd
from nba_api.stats.endpoints import shotchartdetail
from nba_api.stats.static import players

NBA_GAME_ID: Final = "GAME_ID"
NBA_GAME_EVENT_ID: Final = "GAME_EVENT_ID"
NBA_EVENT_TYPE: Final = "EVENT_TYPE"
NBA_PLAYER_NAME: Final = "PLAYER_NAME"
NBA_LOC_X: Final = "LOC_X"
NBA_LOC_Y: Final = "LOC_Y"
NBA_SHOT_DISTANCE: Final = "SHOT_DISTANCE"
NBA_SHOT_TYPE: Final = "SHOT_TYPE"
NBA_ACTION_TYPE: Final = "ACTION_TYPE"
NBA_SHOT_ZONE_AREA: Final = "SHOT_ZONE_AREA"
NBA_SHOT_ZONE_BASIC: Final = "SHOT_ZONE_BASIC"


def find_nba_player(
    player_name: str,
    raise_error_not_unique: bool = True,
) -> List[int]:
    """
    Finds the player identifiers associated to a name.

    :param player_name: Player's name.
    :param raise_error_not_unique: Whether to raise an exception if more than
     one identifier is found.
    :return: A list of player identifiers corresponding to the input name.
    """
    result_players = players.find_players_by_full_name(player_name)

    if raise_error_not_unique and len(result_players) > 1:
        raise ValueError("More than one player found")

    return [player["id"] for player in result_players]


def download_nba_shots(
    season: str, player_name: str = None, player_id: int = None
) -> pd.DataFrame:
    """
    Downloads NBA shot data from NBA's API.

    :param season: Regular season when the shots were made.
    :param player_name: Player whose shots are to be downloaded.
    :param player_id: Player's unique identifier.
    :return: A data frame with shot data.
    """

    if player_id is None:
        if player_name is None:
            raise ValueError(
                "Either a player id. or a player name must be provided"
            )

        player_id = find_nba_player(
            player_name=player_name, raise_error_not_unique=True
        )

    # Query player shots.

    response = shotchartdetail.ShotChartDetail(
        team_id=0,
        player_id=player_id,
        season_nullable=season,
        season_type_all_star="Regular Season",
    )

    content = json.loads(response.get_json())

    # Build data frame.

    results = content["resultSets"][0]
    headers = results["headers"]
    rows = results["rowSet"]

    df = pd.DataFrame(rows)
    df.columns = headers

    # Select relevant columns.

    return df[
        [
            NBA_GAME_ID,
            NBA_GAME_EVENT_ID,
            NBA_EVENT_TYPE,
            NBA_PLAYER_NAME,
            NBA_LOC_X,
            NBA_LOC_Y,
            NBA_SHOT_DISTANCE,
            NBA_SHOT_TYPE,
            NBA_ACTION_TYPE,
            NBA_SHOT_ZONE_AREA,
            NBA_SHOT_ZONE_BASIC,
        ]
    ]


def save_nba_shots(
    season: str, csv: str, player_name: str = None, player_id: int = None
):
    """
    Downloads NBA shot data from NBA's API and saves it in a CSV file.

    :param season: Regular season when the shots were made.
    :param csv: Name of the CSV where data should be saved.
    :param player_name: Player whose shots are to be downloaded.
    :param player_id: Player's unique identifier.
    """
    df = download_nba_shots(
        player_name=player_name, player_id=player_id, season=season
    )
    df.to_csv(csv, index=False)


def _config_args_parser() -> argparse.ArgumentParser:
    """
    Builds and returns the arguments parser for this program.

    :return: The arguments parser for this program.
    """
    parser = argparse.ArgumentParser(
        description="Downloads shot data from an NBA player into a CSV file"
    )

    parser.add_argument(
        "--player-name",
        type=str,
        help="Player name",
    )

    parser.add_argument(
        "--player-id",
        type=int,
        help="Player identifier",
    )

    parser.add_argument(
        "--season",
        type=str,
        required=True,
        help="Season code",
    )

    parser.add_argument(
        "--csv",
        type=str,
        required=True,
        help="CSV file",
    )

    return parser


def main():
    """
    Application entry point.
    """

    # Parse args.

    parser = _config_args_parser()
    args = parser.parse_args(sys.argv[1:])

    # Download CSV.

    save_nba_shots(
        player_name=args.player_name,
        player_id=args.player_id,
        season=args.season,
        csv=args.csv,
    )


if __name__ == "__main__":
    main()
