"""
NBA applications of KDE features.
"""
import os.path
from typing import Final, List

import numpy as np
from matplotlib.image import imread
from pkg_resources import resource_filename

from kdfeatures.density import KDE
from kdfeatures.plot import (
    PDFFeature,
    PDFSignedFeature,
    PDFUnsignedHeight,
    plot_confidence_sets_2d,
    plot_sample_2d,
    plot_surface_2d,
)

_NBA_COURT_PATH: Final = resource_filename(
    __name__, os.path.join("jpg", "nba_court.jpg")
)
"""
NBA court image path.
Courtesy of https://datavizardry.com/2020/02/03/nba-shot-charts-part-2/.
"""

_NBA_COURT_X_MIN: Final = -250
_NBA_COURT_X_MAX: Final = 250
_NBA_COURT_Y_MIN: Final = -52
_NBA_COURT_Y_MAX: Final = 418


def plot_scatter_shots(
    kde: KDE,
    contour_features: List[PDFSignedFeature],
    colors: List[str],
    scatter_feature: PDFFeature = None,
    linestyles: List[str] = None,
    numx: int = 101,
    numy: int = 101,
    scatter_size: float = 10.0,
    contour_linewidth: float = 2.0,
    contour_opacity: float = 1.0,
    fill_contour: bool = False,
):
    """
    Plots a sample of NBA shots along with the fitted KDE features.

    :param kde: Kernel density estimate of the NBA shots.
    :param contour_features: pdf features to extract zero-level contours.
    :param colors: Colors for the contours.
    :param scatter_feature: Feature to color scatter dots.
    :param linestyles: Line styles for the contours.
    :param numx: Number of points in the x axis.
    :param numy: Number of points in the y axis.
    :param scatter_size: Size of observation markers.
    :param contour_linewidth: Width of feature countours.
    :param contour_opacity: A number between 0 (transparent) and 1 (opaque).
    :param fill_contour: Whether to fill the contour. Only used when a single
     contour is present.
    """

    # Plot KDE sample, pdf and features.

    ax = plot_sample_2d(
        pdf=kde,
        contour_features=contour_features,
        scatter_feature=scatter_feature,
        contour_colors=colors,
        contour_linestyles=linestyles,
        xmin=_NBA_COURT_X_MIN,
        xmax=_NBA_COURT_X_MAX,
        numx=numx,
        ymin=_NBA_COURT_Y_MIN,
        ymax=_NBA_COURT_Y_MAX,
        numy=numy,
        sample=kde.underlying_sample,
        scatter_size=scatter_size,
        contour_linewidth=contour_linewidth,
        contour_opacity=contour_opacity,
        fill_contour=fill_contour,
    )

    # Plot background NBA court image.

    _show_nba_court(ax)

    # Rotate.

    ax.invert_yaxis()

    # Remove ticks.

    ax.set_xticks([])
    ax.set_yticks([])


def plot_confidence_shots(
    kde: KDE,
    confidence: float,
    num_bootstrap_samples: int,
    feature: PDFSignedFeature,
    numx: int = 101,
    numy: int = 101,
    scatter_size: float = 10.0,
    contour_opacity: float = 0.5,
    rand: np.random.RandomState = None,
):
    """
    Plots confidence sets for NBA shot data.

    :param kde: Kernel density estimate of the NBA shots.
    :param confidence: Desired probability of the true bump being comprised
     between the confidence sets.
    :param num_bootstrap_samples: Number of bootstrap samples.
    :param feature: Curvature feature corresponding to the requested bump.
    :param numx: Number of points in the x axis.
    :param numy: Number of points in the y axis.
    :param scatter_size: Size of observation markers.
    :param contour_opacity: A number between 0 (transparent) and 1 (opaque)
     for confidence sets.
    :param rand: Random generator.
    """
    # Plot KDE sample and curvature confidence regions.

    ax = plot_confidence_sets_2d(
        kde=kde,
        confidence=confidence,
        num_bootstrap_samples=num_bootstrap_samples,
        feature=feature,
        xmin=_NBA_COURT_X_MIN,
        xmax=_NBA_COURT_X_MAX,
        numx=numx,
        ymin=_NBA_COURT_Y_MIN,
        ymax=_NBA_COURT_Y_MAX,
        numy=numy,
        rand=rand,
        scatter_size=scatter_size,
        contour_opacity=contour_opacity,
    )

    # Plot background NBA court image.

    _show_nba_court(ax)

    # Rotate.

    ax.invert_yaxis()

    # Remove ticks.

    ax.set_xticks([])
    ax.set_yticks([])


def plot_surface_shots(
    kde: KDE,
    numx: int,
    numy: int,
    xstride: int = 1,
    ystride: int = 1,
    area_feature: PDFFeature = PDFUnsignedHeight(),
):
    """
    Plots NBA shots KDE density and one of its features.

    :param kde: Kernel density estimate of the NBA shots.
    :param numx: Number of points in the x axis.
    :param numy: Number of points in the y axis.
    :param xstride: Sampling rate for graphing in the x axis.
    :param ystride: Sampling rate for graphing in the y axis.
    :param area_feature: Density feature to plot as colors over the
     surface area.
    """
    ax = plot_surface_2d(
        pdf=kde,
        area_feature=area_feature,
        xmin=_NBA_COURT_X_MIN,
        xmax=_NBA_COURT_X_MAX,
        ymin=_NBA_COURT_Y_MIN,
        ymax=_NBA_COURT_Y_MAX,
        numx=numx,
        numy=numy,
        xstride=xstride,
        ystride=ystride,
        elevation=30.0,
        azimuth=120.0,
    )

    # NBA court image.

    arr = imread(_NBA_COURT_PATH) / 255

    x = np.flip(
        np.linspace(
            start=_NBA_COURT_X_MIN, stop=_NBA_COURT_X_MAX, num=arr.shape[1]
        )
    )
    y = np.flip(
        np.linspace(
            start=_NBA_COURT_Y_MIN, stop=_NBA_COURT_Y_MAX, num=arr.shape[0]
        )
    )
    x, y = np.meshgrid(x, y)

    ax.plot_surface(
        x,
        y,
        ax.get_zlim()[1] * np.ones(x.shape),
        rstride=1,
        cstride=1,
        facecolors=arr,
        alpha=0.05,
        shade=False,
    )

    # Rotate.

    ax.invert_xaxis()

    # Remove ticks.

    ax.set_xticks([])
    ax.set_yticks([])
    ax.set_zticks([])


def _show_nba_court(ax):
    """
    Plots a background image of an NBA court.

    :param ax: Axes where to plot the NBA court.
    """
    ax.imshow(
        imread(_NBA_COURT_PATH),
        extent=[
            _NBA_COURT_X_MIN,
            _NBA_COURT_X_MAX,
            _NBA_COURT_Y_MIN,
            _NBA_COURT_Y_MAX,
        ],
    )
