"""
Access to downloaded NBA shot data.
"""
import os.path
from typing import Final

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from kdfeatures.applications.nba.api import NBA_LOC_X, NBA_LOC_Y

_KEVIN_DURANT_SHOTS_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "kevin_durant_2015_16.csv")
)

_STEPHEN_CURRY_SHOTS_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "stephen_curry_2015_16.csv")
)

_JAMES_HARDEN_SHOTS_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "james_harden_2015_16.csv")
)


def read_shots(csv: str, unique: bool = False) -> np.ndarray:
    """
    Reads shot data from a CSV file.

    :param csv: CSV to read data from.
    :param unique: Whether to return only non-duplicate shot coordinates.
    :return: A matrix where each row corresponds to a pair of shot coordinates.
    """
    column_names: Final = [NBA_LOC_X, NBA_LOC_Y]

    df: Final = pd.read_csv(csv, usecols=column_names)

    columns: Final = [np.array(df[name]).astype(float) for name in column_names]

    shots = np.column_stack(tup=tuple(columns))

    if unique:
        shots = np.unique(shots, axis=0)

    return shots


def kevin_durant_shots(unique: bool = False) -> np.ndarray:
    """
    Returns shots by Kevin Durant in the 2015-16 season.

    :param unique: Whether to return only non-duplicate shot coordinates.
    :return: Shots by Kevin Durant.
    """
    return read_shots(csv=_KEVIN_DURANT_SHOTS_CSV_PATH, unique=unique)


def stephen_curry_shots(unique: bool = False) -> np.ndarray:
    """
    Returns shots by Stephen Curry in the 2015-16 season.

    :param unique: Whether to return only non-duplicate shot coordinates.
    :return: Shots by Stephen Curry.
    """
    return read_shots(csv=_STEPHEN_CURRY_SHOTS_CSV_PATH, unique=unique)


def james_harden_shots(unique: bool = False) -> np.ndarray:
    """
    Returns shots by James Harden in the 2015-16 season.

    :param unique: Whether to return only non-duplicate shot coordinates.
    :return: Shots by James Harden.
    """
    return read_shots(csv=_JAMES_HARDEN_SHOTS_CSV_PATH, unique=unique)
