"""
Access NFL data from the API.
"""
import argparse
import sys
from typing import Final, List

import nflgame
import pandas as pd

NFL_GAME_KEY: Final = "gamekey"
NFL_DRIVE_NUM: Final = "drive_num"
NFL_PLAY_ID: Final = "playid"
NFL_PASSER_ID: Final = "passing_cmp.playerid"
NFL_PASSER_NAME: Final = "passing_cmp.playername"
NFL_PASSING_YARDS: Final = "passing_yds"
NFL_RECEIVER_ID: Final = "receiving_tar.playerid"
NFL_RECEIVER_NAME: Final = "receiving_tar.playername"


def find_nfl_player(
    player_name: str, raise_error_not_unique: bool = True
) -> List[str]:
    """
    Finds the player identifiers associated to a name.

    :param player_name: Player's name.
    :param raise_error_not_unique: Whether to raise an exception if more than
     one identifier is found.
    :return: A list of player identifiers corresponding to the input name.
    """
    result_players = nflgame.find(name=player_name)

    if raise_error_not_unique and len(result_players) > 1:
        raise ValueError("More than one player found")

    return [player.player_id for player in result_players]


def nfl_completions(
    year: int, player_name: str = None, player_id: str = None
) -> pd.DataFrame:
    """
    Obtains a data frame of pass completions.

    :param year: NFL season.
    :param player_name: Name of the passer.
    :param player_id: Player's unique identifier.
    :return: A data frame comprising pass completions at the play level, with
     information about the passer, the receiver and total passing yards.
    """
    if player_id is None:
        if player_name is None:
            raise ValueError(
                "Either a player id. or a player name must be provided"
            )

        player_id = find_nfl_player(
            player_name=player_name, raise_error_not_unique=True
        )[0]

    player: Final = nflgame.players[player_id]
    completions: Final = player.plays(year=year).filter(
        passing_cmp=1, receiving_tar=1
    )

    play_dicts = []

    for play in completions:
        play_dict = {}

        passer_id, passer_name = next(
            (event["playerid"], event["playername"])
            for event in play.events
            if "passing_cmp_air_yds" in event
        )

        receiver_id, receiver_name = next(
            (event["playerid"], event["playername"])
            for event in play.events
            if "receiving_tar" in event
        )

        play_dict[NFL_GAME_KEY] = play.drive.game.gamekey
        play_dict[NFL_DRIVE_NUM] = play.drive.drive_num
        play_dict[NFL_PLAY_ID] = play.playid
        play_dict[NFL_PASSER_ID] = passer_id
        play_dict[NFL_PASSER_NAME] = passer_name
        play_dict[NFL_RECEIVER_ID] = receiver_id
        play_dict[NFL_RECEIVER_NAME] = receiver_name
        play_dict[NFL_PASSING_YARDS] = play.passing_yds

        play_dicts.append(play_dict)

    return pd.DataFrame(
        play_dicts,
        columns=[
            NFL_GAME_KEY,
            NFL_DRIVE_NUM,
            NFL_PLAY_ID,
            NFL_PASSER_ID,
            NFL_PASSER_NAME,
            NFL_RECEIVER_ID,
            NFL_RECEIVER_NAME,
            NFL_PASSING_YARDS,
        ],
    )


def _config_args_parser() -> argparse.ArgumentParser:
    """
    Builds and returns the arguments parser for this program.

    :return: The arguments parser for this program.
    """
    parser = argparse.ArgumentParser(
        description="Downloads shot data from an NBA player into a CSV file"
    )

    parser.add_argument(
        "--player-name",
        type=str,
        help="Player name",
    )

    parser.add_argument(
        "--player-id",
        type=str,
        help="Player identifier",
    )

    parser.add_argument(
        "--year",
        type=int,
        required=True,
        help="Year",
    )

    parser.add_argument(
        "--csv",
        type=str,
        required=True,
        help="CSV file",
    )

    return parser


def save_nfl_completions(
    year: int, csv: str, player_name: str = None, player_id: str = None
):
    """
    Saves NFL completions data into a CSV file.

    :param year: NFL season.
    :param csv: Name of the CSV where data should be saved.
    :param player_name: Name of the passer.
    :param player_id: Player's unique identifier.
    """
    df = nfl_completions(
        player_name=player_name, player_id=player_id, year=year
    )

    df.to_csv(csv, index=False)


def main():
    """
    Application entry point.
    """

    # Parse args.

    parser = _config_args_parser()
    args = parser.parse_args(sys.argv[1:])

    # Download CSV.

    save_nfl_completions(
        player_name=args.player_name,
        player_id=args.player_id,
        year=args.year,
        csv=args.csv,
    )


if __name__ == "__main__":
    main()
