"""
NFL applications of KDE features.
"""
import os
from typing import Final

from matplotlib import pyplot as plt
from pkg_resources import resource_filename

from kdfeatures.density import KDE
from kdfeatures.plot import PDFFeature, PDFSignedFeature, plot_sample_1d

_NFL_FIELD_PATH: Final = resource_filename(
    __name__, os.path.join("png", "nfl_field.png")
)
"""
NFL field image path.
From https://en.wikipedia.org/wiki/American_football_field.
"""


def plot_scatter_yards(
    kde: KDE,
    contour_feature: PDFSignedFeature = None,
    scatter_feature: PDFFeature = None,
    min_yards: int = -10,
    max_yards: int = 145,
    aspect_multiplier: float = 1.2,
    colorbar_shrink: float = 0.5,
    **kwargs,
):
    """
    Plots scatter passing yards data.

    :param kde: Kernel density estimate of the data.
    :param contour_feature: pdf feature to extract the zero-level set.
    :param scatter_feature: Feature to color scatter dots.
    :param min_yards: Minimum horizontal axis value.
    :param max_yards: Maximum horizontal axis value.
    :param aspect_multiplier: Factor that expands or compresses the NFL field
     background image.
    :param colorbar_shrink: Shrink factor for color bar.
    :param kwargs: Keyword arguments for sample 1D scatter plot.
    """
    # Plot NFL field as background image.

    _, ax = plt.subplots()

    nfl_field_image = plt.imread(_NFL_FIELD_PATH)

    extent = [-10, 110, 0, 1]
    ax.imshow(nfl_field_image, extent=[-130, -10, 0, 1])
    ax.imshow(nfl_field_image, extent=extent)
    ax.imshow(nfl_field_image, extent=[110, 230, 0, 1])
    ax.set_xlim(min_yards, max_yards)

    aspect = (
        nfl_field_image.shape[0]
        / float(nfl_field_image.shape[1])
        * ((extent[1] - extent[0]) / (extent[3] - extent[2]))
    )

    plt.gca().set_aspect(aspect_multiplier * aspect)

    # Plot yards scatter plot.

    plot_sample_1d(
        pdf=kde,
        sample=kde.underlying_sample,
        contour_feature=contour_feature,
        scatter_feature=scatter_feature,
        ax=ax,
        colorbar_shrink=colorbar_shrink,
        **kwargs,
    )
