"""
Access to saved NFL completion data.
"""
import os
from typing import Final

import numpy as np
import pandas as pd
from pkg_resources import resource_filename

from kdfeatures.applications.nfl.api import (
    NFL_GAME_KEY,
    NFL_PASSING_YARDS,
    NFL_RECEIVER_ID,
)

_TOM_BRADY_2014_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "tom_brady_2014.csv")
)

_TOM_BRADY_2016_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "tom_brady_2016.csv")
)

_TOM_BRADY_2018_CSV_PATH: Final = resource_filename(
    __name__, os.path.join("csv", "tom_brady_2018.csv")
)


def sum_passing_yards_by_appearance(completions: pd.DataFrame) -> np.ndarray:
    """
    Aggregates a completions data frame by summing passing yards grouped by
    game and receiver.

    :param completions: Data frame with NFL passing completions.
    :return: An array where each element corresponds to the total passing yards
     of a receiver appearance.
    """
    passing_yds_player_game = (
        completions[
            [
                NFL_GAME_KEY,
                NFL_RECEIVER_ID,
                NFL_PASSING_YARDS,
            ]
        ]
        .groupby(
            by=[
                NFL_GAME_KEY,
                NFL_RECEIVER_ID,
            ],
        )
        .sum()
    )

    return np.array(passing_yds_player_game.passing_yds)


def passing_yards_by_appearance(csv: str) -> np.ndarray:
    """
    Obtains a sample of passing yards per appearance from a CSV with
    completions data.

    :param csv: CSV file to read.
    :return: An array with passing yards per individual appearance.
    """
    return sum_passing_yards_by_appearance(
        pd.read_csv(
            csv, usecols=[NFL_GAME_KEY, NFL_RECEIVER_ID, NFL_PASSING_YARDS]
        )
    )


def passing_yards_by_tom_brady_2014() -> np.ndarray:
    """
    Returns an array with the individual passing yard records arising from
    Tom Brady in the 2014 season.

    :return: An array of individual passing yards per appearance in the 2014
     season arising from Tom Brady's passes.
    """
    return passing_yards_by_appearance(_TOM_BRADY_2014_CSV_PATH)


def passing_yards_by_tom_brady_2016() -> np.ndarray:
    """
    Returns an array with the individual passing yard records arising from
    Tom Brady in the 2016 season.

    :return: An array of individual passing yards per appearance in the 2016
     season arising from Tom Brady's passes.
    """
    return passing_yards_by_appearance(_TOM_BRADY_2016_CSV_PATH)


def passing_yards_by_tom_brady_2018() -> np.ndarray:
    """
    Returns an array with the individual passing yard records arising from
    Tom Brady in the 2018 season.

    :return: An array of individual passing yards per appearance in the 2018
     season arising from Tom Brady's passes.
    """
    return passing_yards_by_appearance(_TOM_BRADY_2018_CSV_PATH)
