"""
Gaussian bandwidth selection.
"""
from typing import Callable, Final

import numpy as np
from numpy.linalg import inv
from sklearn.mixture import GaussianMixture

from kdfeatures.bandwidth.optim import (
    Bandwidths,
    bandwidth_matrix_to_scalar,
    optimize_bandwidth,
    scalar_bandwidth_matrix,
)
from kdfeatures.density import Gaussian
from kdfeatures.util.hermite import hermite_minimal_polynomial
from kdfeatures.util.matrix import is_matrix


def gaussian_bandwidth(sample: np.ndarray, r: int) -> np.ndarray:
    """
    Chooses the bandwidth parameters assuming the underlying function is
    Gaussian.

    :param sample: Sample to estimate bandwidths from.
    :param r: Target differentiation order.
    :return: The estimated bandwidth matrix.
    """
    if not is_matrix(sample):
        raise ValueError("Sample is not a matrix")

    n: Final = sample.shape[0]
    d: Final = sample.shape[1]
    cov: Final = np.cov(sample, rowvar=False).reshape((d, d))

    return (4 / (n * (d + 2 * r + 2))) ** (2 / (d + 2 * r + 4)) * cov


def univariate_gaussian_bandwidth(sample: np.ndarray, r: int) -> float:
    """
    Chooses the bandwidth parameter assuming the underlying function is
    Gaussian.

    :param sample: Sample to estimate bandwidths from.
    :param r: Target differentiation order.
    :return: The estimated bandwidth parameter.
    """
    return float(
        np.sqrt(
            gaussian_bandwidth(sample=np.expand_dims(a=sample, axis=1), r=r)[
                0, 0
            ]
        )
    )


def _mixture_component_derivatives(
    derivatives_order: int,
    means: np.ndarray,
    covariances: np.ndarray,
) -> np.ndarray:
    """
    Calculates Gaussian derivatives for the Gaussian mixture bandwidth selector.

    :param derivatives_order: Order of derivatives.
    :param means: Gaussian mixture means.
    :param covariances: Gaussian mixture covariance matrices.
    :return: A square matrix of derivative vectors.
    """

    num_components: Final = len(means)
    assert len(covariances) == num_components

    dimension: Final = len(means[0])
    derivatives: Final = np.empty(
        (num_components, num_components, dimension**derivatives_order)
    )

    for i in range(num_components):
        for j in range(num_components):
            gaussian = Gaussian(
                dimension=dimension, covariance=covariances[i] + covariances[j]
            )

            derivatives[i, j, :] = gaussian.derivative(
                r=derivatives_order,
                x=np.expand_dims(means[i] - means[j], axis=0),
            )

    return derivatives


def gaussian_mixture_loss_function(
    sample: np.ndarray,
    r: int,
    num_components: int,
    rand: np.random.RandomState = None,
) -> Callable[[np.ndarray], float]:
    """
    Builds the loss function for selecting the optimal Gaussian mixture
    bandwidth.

    :param sample: Sample to estimate the bandwidth from.
    :param r: Differentiation order.
    :param num_components: Number of mixture components.
    :param rand: Random generator state for Gaussian mixture initialization.
    :return: The loss function for the Gaussian mixture bandwidth selection.
    """

    # Shape parameters.

    n: Final = len(sample)
    d: Final = sample.shape[1]
    twice_r: Final = 2 * r

    # Hermite polynomial.

    hermite_poly: Final = np.squeeze(
        hermite_minimal_polynomial(
            x=np.zeros((1, d)), v=-np.eye(d), order=twice_r
        ).enlarged_vector()
    ).reshape(-1)
    assert len(hermite_poly) == d**twice_r

    # Fit Gaussian mixture.

    gaussian_mixture: Final = GaussianMixture(
        n_components=num_components, random_state=rand
    )
    gaussian_mixture.fit(X=sample)

    w: Final = gaussian_mixture.weights_.reshape((1, -1))

    # Calculate auxiliary derivatives.

    derivatives: Final = _mixture_component_derivatives(
        derivatives_order=twice_r + 4,
        means=gaussian_mixture.means_,
        covariances=gaussian_mixture.covariances_,
    )

    # Other auxiliary terms.

    sign: Final = (-1) ** r
    id_r: Final = np.eye(d**r)

    k: Final = (2 ** (-r)) * (4 * np.pi ** (-d / 2)) / n

    # Define loss function.

    def nm(h: np.ndarray) -> float:
        vec_inv_h: Final = inv(h).flatten()
        vec_inv_h_r = 1

        for _ in range(r):
            vec_inv_h_r = np.kron(vec_inv_h_r, vec_inv_h)

        vec_h: Final = h.flatten()
        outer_vec_h: Final = np.outer(vec_h, vec_h)

        v: Final = np.kron(id_r, outer_vec_h).flatten()
        m: Final = np.empty((num_components, num_components))

        for i in range(num_components):
            for j in range(num_components):
                m[i, j] = np.inner(v, derivatives[i, j])

        return float(
            k
            * float(np.linalg.det(h) ** (-1 / 2))
            * float(np.dot(vec_inv_h_r, hermite_poly))
            + sign * (np.dot(w, np.dot(m, w.T))) / 4
        )

    return nm


def gaussian_mixture_bandwidth(
    sample: np.ndarray,
    r: int,
    num_components: int,
    bandwidth_type: Bandwidths = Bandwidths.unconstrained,
    rand: np.random.RandomState = None,
    print_progress: bool = False,
    print_loss: bool = False,
    print_every_num_iter: int = 1,
    **options,
) -> np.ndarray:
    """
    Calculates the Gaussian mixture bandwidth.

    :param sample: Sample to estimate the bandwidth from.
    :param r: Differentiation order.

    :param num_components: Number of mixture components.
    :param bandwidth_type: Type of bandwidth (scalar, unconstrained, etc.).
    :param rand: Random generator state for Gaussian mixture initialization.
    :param print_progress: Whether to print optimization progress with each
     iteration.
    :param print_loss: Whether to print the loss value with each iteration.
    :param print_every_num_iter: Print intermediate results every number of
     iterations according to this argument.
    :param options: Keyword arguments for the internal optimizer.
    :return: The optimal bandwidth matrix.
    """

    h0 = gaussian_bandwidth(sample=sample, r=r)

    if bandwidth_type is Bandwidths.scalar:
        h0 = bandwidth_matrix_to_scalar(h0)

    dimension: Final = sample.shape[1]

    def build_bandwidth(h):
        return (
            scalar_bandwidth_matrix(h=h, dimension=dimension)
            if bandwidth_type is Bandwidths.scalar
            else h
        )

    return build_bandwidth(
        optimize_bandwidth(
            loss_matrix=gaussian_mixture_loss_function(
                sample=sample, r=r, num_components=num_components, rand=rand
            ),
            h0=h0,
            dimension=dimension,
            print_progress=print_progress,
            print_loss=print_loss,
            print_every_num_iter=print_every_num_iter,
            **options,
        )
    )
