"""
General optimization of bandwidth matrices.
"""
from enum import Enum, auto
from typing import Callable, Final, Union

import numpy as np
from scipy.optimize import minimize

from kdfeatures.util.matrix import from_log_cholesky, to_log_cholesky

_num_iter = 0

# noinspection PyArgumentList
class Bandwidths(Enum):
    """
    Types of bandwidth matrices.
    """

    unconstrained = auto()
    scalar = auto()


def scalar_bandwidth_matrix(h: float, dimension: int) -> np.ndarray:
    """
    Builds a scalar bandwidth matrix.

    :param h: Scalar value.
    :param dimension: Dimension of the matrix.
    :return: The corresponding scalar bandwidth matrix.
    """
    return (h**2) * np.identity(dimension)


def bandwidth_matrix_to_scalar(h: np.ndarray) -> float:
    """
    Builds a scalar bandwidth from a bandwidth matrix.

    :param h: Bandwidth matrix.
    :return: A scalar bandwidth that roughly provides the same smoothing as the
     input matrix.
    """
    return float(np.linalg.det(h) ** (1 / (2 * h.shape[0])))


def optimize_bandwidth(
    loss_matrix: Callable[[np.ndarray], float],
    h0: Union[np.ndarray, float],
    dimension: int = None,
    print_progress: bool = False,
    print_loss: bool = False,
    print_every_num_iter: int = 1,
    **options,
) -> np.ndarray:
    """
    Optimizes a loss function with a bandwidth matrix input.

    :param loss_matrix: Loss function to minimize.
    :param h0: Initial bandwidth.
    :param dimension: Dimension of the target bandwidth matrix. Only needed for
     scalar bandwidths.
    :param print_progress: Whether to print optimization progress with each
     iteration.
    :param print_loss: Whether to print the loss value with each iteration.
    :param print_every_num_iter: Print intermediate results every number of
     iterations according to this argument.
    :param options: Keyword arguments for the internal optimizer.
    :return: The optimal bandwidth matrix.
    """
    if isinstance(h0, float) and dimension is None:
        raise ValueError("Missing bandwidth dimension")

    matrix_from_params: Final = (
        from_log_cholesky
        if isinstance(h0, np.ndarray)
        else lambda x: np.exp(x[0]) * np.identity(dimension)
    )

    loss_vector: Final = lambda x: loss_matrix(matrix_from_params(x))

    def print_h(x):
        global _num_iter

        if _num_iter % print_every_num_iter == 0:
            h = matrix_from_params(x)

            with np.printoptions(precision=10, suppress=True):
                print(f"> H_{_num_iter} =\n" + str(h))

                if print_loss:
                    print(f"> loss(H_{_num_iter}) = " + str(loss_matrix(h)))

        _num_iter += 1

    x0: Final = (
        to_log_cholesky(h0)
        if isinstance(h0, np.ndarray)
        else np.array([2 * np.log(h0)])
    )

    if print_progress:
        print_h(x0)

    result: Final = minimize(
        fun=loss_vector,
        x0=x0,
        method="Nelder-Mead",
        callback=print_h if print_progress else None,
        options=options,
    )

    if print_progress:
        global _num_iter
        _num_iter = 0

    bandwidth = matrix_from_params(result.x)

    return bandwidth if isinstance(h0, np.ndarray) else np.sqrt(bandwidth[0, 0])
