"""
Smoothed cross validation.
"""
from typing import Callable, Final

import numpy as np
from numpy.linalg import inv

from kdfeatures.bandwidth import (
    gaussian_bandwidth,
)
from kdfeatures.bandwidth.optim import (
    Bandwidths,
    bandwidth_matrix_to_scalar,
    optimize_bandwidth,
    scalar_bandwidth_matrix,
)
from kdfeatures.density import Gaussian
from kdfeatures.util.batch import double_sum_batches
from kdfeatures.util.hermite import (
    hermite_minimal_polynomial,
)


def scv_loss_function(
    sample: np.ndarray, r: int, pilot_g: np.ndarray, num_batches: int = None
) -> Callable[[np.ndarray], float]:
    """
    Builds the loss function for smoothed cross validation (SCV) bandwidth
    estimation.

    :param sample: Random sample to base loss value calculations upon.
    :param pilot_g: Pilot bandwidth matrix.
    :param r: Derivative order.
    :param num_batches: Number of splits for the double sum.
    :return: The loss function for SCV bandwidth estimation.
    """
    n: Final = len(sample)
    n_square: Final = n**2

    twice_r: Final = 2 * r
    minus_one_pow_r: Final = (-1) ** r

    d: Final = sample.shape[1]

    vec_id_r: Final = np.eye(d**r).flatten().reshape((-1, 1))
    assert len(vec_id_r) == d**twice_r

    hermite_poly: Final = np.squeeze(
        hermite_minimal_polynomial(
            x=np.zeros((1, d)), v=-np.eye(d), order=twice_r
        ).enlarged_vector()
    ).reshape(-1)
    assert len(hermite_poly) == d**twice_r

    k: Final = (2 ** (-(d + r))) * (np.pi ** (-d / 2)) / n
    twice_pilog_g: Final = 2 * pilot_g

    def sum_cv(gaussian: Gaussian) -> float:
        sum_ = 0

        for batch in double_sum_batches(sample=sample, num_batches=num_batches):
            sum_ += float(
                np.sum(
                    np.dot(gaussian.derivative(x=batch, r=twice_r), vec_id_r)
                )
            )

        return sum_

    def scv(h: np.ndarray) -> float:
        vec_inv_h: Final = inv(h).flatten()
        vec_inv_h_r = 1

        for _ in range(r):
            vec_inv_h_r = np.kron(vec_inv_h_r, vec_inv_h)

        gaussian_2h_2g: Final = Gaussian(
            dimension=d, covariance=2 * h + twice_pilog_g
        )
        gaussian_h_2g: Final = Gaussian(
            dimension=d, covariance=h + twice_pilog_g
        )
        gaussian_2g: Final = Gaussian(dimension=d, covariance=twice_pilog_g)

        return (
            k
            * float(np.linalg.det(h) ** (-1 / 2))
            * float(np.dot(vec_inv_h_r, hermite_poly))
            + minus_one_pow_r
            * (
                sum_cv(gaussian=gaussian_2h_2g)
                - 2 * sum_cv(gaussian=gaussian_h_2g)
                + sum_cv(gaussian=gaussian_2g)
            )
            / n_square
        )

    return scv


def smoothed_cv_bandwidth(
    sample: np.ndarray,
    r: int,
    bandwidth_type: Bandwidths = Bandwidths.unconstrained,
    pilot_g: np.ndarray = None,
    num_stages: int = 1,
    num_batches: int = None,
    print_progress: bool = False,
    print_loss: bool = False,
    print_every_num_iter: int = 1,
    **options,
) -> np.ndarray:
    """
    Calculates the smoothed cross validation bandwidth.

    :param sample: Sample to estimate the bandwidth from.
    :param r: Differentiation order.
    :param bandwidth_type: Type of bandwidth (scalar, unconstrained, etc.).
    :param pilot_g: Pilot bandwidth matrix.
    :param num_stages: Number of smoothing iterations.
    :param num_batches: Number of batches for computing the double sum.
    :param print_progress: Whether to print optimization progress with each
     iteration.
    :param print_loss: Whether to print the loss value with each iteration.
    :param print_every_num_iter: Print intermediate results every number of
     iterations according to this argument.
    :param options: Keyword arguments for the internal optimizer.
    :return: The optimal bandwidth matrix.
    """
    if pilot_g is None:
        pilot_g = gaussian_bandwidth(sample=sample, r=r)

    dimension: Final = sample.shape[1]

    if bandwidth_type is Bandwidths.unconstrained:
        h = pilot_g
    else:
        h = bandwidth_matrix_to_scalar(pilot_g)

    def build_bandwidth(x):
        return (
            scalar_bandwidth_matrix(h=x, dimension=dimension)
            if bandwidth_type is Bandwidths.scalar
            else x
        )

    for i in range(1, num_stages + 1):
        if print_progress:
            print(f"*** Stage {i} ***")

        h = optimize_bandwidth(
            loss_matrix=scv_loss_function(
                sample=sample,
                r=r,
                pilot_g=build_bandwidth(h),
                num_batches=num_batches,
            ),
            h0=h,
            dimension=dimension,
            print_progress=print_progress,
            print_loss=print_loss,
            print_every_num_iter=print_every_num_iter,
            **options,
        )

    return build_bandwidth(h)


def univariate_smoothed_cv_bandwidth(
    sample: np.ndarray,
    r: int,
    pilot_g: float = None,
    num_stages: int = 1,
    num_batches: int = None,
    print_progress: bool = False,
    print_loss: bool = False,
    print_every_num_iter: int = 1,
    **options,
) -> float:
    """
    Calculates the smoothed cross validation bandwidth for a univariate
    sample.

    :param sample: Sample to estimate the bandwidth from.
    :param r: Differentiation order.
    :param pilot_g: Pilot bandwidth.
    :param num_stages: Number of smoothing iterations.
    :param num_batches: Number of batches for computing the double sum.
    :param print_progress: Whether to print optimization progress with each
     iteration.
    :param print_loss: Whether to print the loss value with each iteration.
    :param print_every_num_iter: Print intermediate results every number of
     iterations according to this argument.
    :param options: Keyword arguments for the internal optimizer.
    :return: The optimal bandwidth matrix.
    """
    return float(
        smoothed_cv_bandwidth(
            sample=np.expand_dims(a=sample, axis=1),
            r=r,
            pilot_g=np.array([[pilot_g**2]]) if pilot_g is not None else None,
            num_stages=num_stages,
            num_batches=num_batches,
            print_progress=print_progress,
            print_loss=print_loss,
            print_every_num_iter=print_every_num_iter,
            **options,
        )[0, 0]
    )
