"""
Unbiased cross validation.
"""
from typing import Callable, Final

import numpy as np

from kdfeatures.bandwidth import gaussian_bandwidth
from kdfeatures.bandwidth.optim import optimize_bandwidth
from kdfeatures.density import Gaussian
from kdfeatures.util.batch import double_sum_batches


def ucv_loss_function(
    sample: np.ndarray, r: int, num_batches: int = None
) -> Callable[[np.ndarray], float]:
    """
    Builds the loss function for unbiased cross validation (UCV) bandwidth
    estimation.

    :param sample: Random sample to base loss value calculations upon.
    :param r: Derivative order.
    :param num_batches: Number of splits for the double sum.
    :return: The loss function for UCV bandwidth estimation.
    """
    n: Final = len(sample)
    n_square: Final = n**2
    n_square_minus_n: Final = n_square - n

    twice_r: Final = 2 * r
    minus_one_pow_r: Final = (-1) ** r

    d: Final = sample.shape[1]
    zero_vector: Final = np.zeros((1, d))

    vec_id_r: Final = np.eye(d**r).flatten().reshape((-1, 1))
    assert len(vec_id_r) == d**twice_r

    def lscv(gaussian: Gaussian, exclude_diagonal: bool) -> float:
        sum_ = 0

        for batch in double_sum_batches(sample=sample, num_batches=num_batches):
            sum_ += float(
                np.sum(
                    np.dot(gaussian.derivative(x=batch, r=twice_r), vec_id_r)
                )
            )

        if exclude_diagonal:
            sum_ -= n * float(
                np.sum(
                    np.dot(
                        gaussian.derivative(x=zero_vector, r=twice_r), vec_id_r
                    )
                )
            )

            sum_ /= n_square_minus_n
        else:
            sum_ /= n_square

        return sum_

    def ucv(h: np.ndarray) -> float:
        gaussian_2h: Final = Gaussian(dimension=d, covariance=2 * h)
        gaussian_h: Final = Gaussian(dimension=d, covariance=h)

        return minus_one_pow_r * (
            lscv(gaussian=gaussian_2h, exclude_diagonal=False)
            - 2 * lscv(gaussian=gaussian_h, exclude_diagonal=True)
        )

    return ucv


def unbiased_cv_bandwidth(
    sample: np.ndarray,
    r: int,
    num_batches: int = None,
    print_progress: bool = False,
    print_loss: bool = False,
    print_every_num_iter: int = 1,
    **options,
) -> np.ndarray:
    """
    Calculates the unbiased cross validation bandwidth.

    :param sample: Sample to estimate the bandwidth from.
    :param r: Differentiation order.
    :param num_batches: Number of batches for computing the double sum.
    :param print_progress: Whether to print optimization progress with each
     iteration.
    :param print_loss: Whether to print the loss value with each iteration.
    :param print_every_num_iter: Print intermediate results every number of
     iterations according to this argument.
    :param options: Keyword arguments for the internal optimizer.
    :return: The optimal bandwidth matrix.
    """
    return optimize_bandwidth(
        loss_matrix=ucv_loss_function(
            sample=sample, r=r, num_batches=num_batches
        ),
        h0=gaussian_bandwidth(sample=sample, r=r),
        print_progress=print_progress,
        print_loss=print_loss,
        print_every_num_iter=print_every_num_iter,
        **options,
    )
