"""
Bootstrap methods.
"""
from typing import Callable, Final, List, Union

import numpy as np


def bootstrap_sample(
    sample: np.ndarray,
    resample_size: int = None,
    rand: np.random.RandomState = None,
) -> np.ndarray:
    """
    Obtains a bootstrap sample (a resample with replacement) from a given
    sample.

    :param sample: Original sample.
    :param resample_size: Number of observations to draw from the resample.
    :param rand: Random generator.
    :return: A resample of the original sample with replacement.
    """

    if rand is None:
        rand = np.random

    if resample_size is None:
        resample_size = len(sample)

    indices: Final = rand.choice(
        a=len(sample), size=resample_size, replace=True
    )

    return sample[indices]


def bootstrap_samples(
    sample: np.ndarray,
    num_samples: int,
    resample_size: int = None,
    rand: np.random.RandomState = None,
):
    """
    Generates bootstrap samples.

    :param sample: Original sample.
    :param num_samples: Number of bootstrap samples.
    :param resample_size: Number of observations in each bootstrap sample.
    :param rand: Random generator.
    :return: Bootstrap samples, one by one in a generator.
    """
    for _ in range(num_samples):
        yield bootstrap_sample(
            sample=sample, resample_size=resample_size, rand=rand
        )


def bootstrap_statistic(
    statistic: Callable[[np.ndarray], Union[float, np.ndarray]],
    sample: np.ndarray,
    num_samples: int,
    resample_size: int = None,
    rand: np.random.RandomState = None,
) -> List[Union[float, np.ndarray]]:
    """
    Calculates a certain statistic over bootstrap samples.

    :param statistic: Statistic to calculate.
    :param sample: Original sample to bootstrap from.
    :param num_samples: Number of bootstrap samples to generate.
    :param resample_size: Number of observations per bootstrap sample.
    :param rand: Random generator.
    :return: A list of values from the statistic with size equal to the number
     of bootstrap samples requested.
    """

    statistic_values: Final = list()

    for resample in bootstrap_samples(sample, num_samples, resample_size, rand):
        statistic_values.append(statistic(resample))

    return statistic_values
