"""
Probability density functions.
"""
from abc import ABC, abstractmethod
from typing import Callable, Final, List, Tuple

import numpy as np
from scipy.integrate import nquad

from kdfeatures.util.matrix import largest_eigenvalue, smallest_eigenvalue


class PDF(ABC):
    """
    An abstract probability density function.
    """

    @property
    @abstractmethod
    def dimension(self) -> int:
        """
        The dimension of the sampling space.

        :return: The dimension of the sampling space.
        """
        pass

    @abstractmethod
    def sample(
        self, size: int, rand: np.random.RandomState = None
    ) -> np.ndarray:
        """
        Draws a random sample from the pdf.

        :param size: Number of observations to draw.
        :param rand: Random generator.
        :return: A matrix where each row represents an observation.
        """
        pass

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the pdf at the input points.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the pdf values of each point.
        """
        self._check_input(x)
        return self._pdf(x)

    def jacobian(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the Jacobian of the pdf at the input points.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: A matrix with as many rows as input points, where each column
         corresponds to one of the partial derivatives.
        """
        self._check_input(x)
        return self._jacobian(x)

    def hessian(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the Hessian of the pdf at the input points.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: A matrix with as many rows as input points, where each column
         corresponds to one of the second-order partial derivatives.
        """
        self._check_input(x)
        return self._hessian(x)

    def derivative(self, r: int, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the r-th order derivative of the pdf at the input points.

        :param r: Order of the derivatives.
        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: A matrix with as many rows as input points, where each column
         corresponds to one of the r-th order partial derivatives.
        """
        self._check_input(x)
        return self._derivative(r=r, x=x)

    def shape(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the shape operator of the pdf at the input points.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: A matrix with as many rows as input points, where each column
         corresponds to one of the shape operator elements.
        """
        jac: Final = self.jacobian(x)
        hess: Final = self.hessian(x)

        kron_jac: Final = np.apply_along_axis(
            func1d=self._kronecker_jacobian,
            axis=1,
            arr=jac,
        )

        sh = np.array(
            [
                np.squeeze(np.dot(h.reshape((1, -1)), j))
                for h, j in zip(hess, kron_jac)
            ]
        )

        if self.dimension == 1:
            sh = np.expand_dims(sh, axis=1)

        return sh

    def largest_hessian_eigenvalue(self, x: np.ndarray) -> np.ndarray:
        """
        Calculates the largest eigenvalue of the Hessian matrix.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the largest eigenvalues of the Hessian matrix at each point.
        """
        return self._hessian_functional(
            x=x, matrix_functional=largest_eigenvalue
        )

    def smallest_hessian_eigenvalue(self, x: np.ndarray) -> np.ndarray:
        """
        Calculates the smallest eigenvalue of the Hessian matrix.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the smallest eigenvalues of the Hessian matrix at each point.
        """
        return self._hessian_functional(
            x=x, matrix_functional=smallest_eigenvalue
        )

    def mean_curvature(self, x: np.ndarray) -> np.ndarray:
        """
        Calculates the mean curvature.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the mean curvatures at each point.
        """
        return self._shape_functional(x=x, matrix_functional=np.trace)

    def gaussian_curvature(self, x: np.ndarray) -> np.ndarray:
        """
        Calculates the Gaussian curvature.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the Gaussian curvatures at each point.
        """
        return self._shape_functional(x=x, matrix_functional=np.linalg.det)

    def laplacian(self, x: np.ndarray) -> np.ndarray:
        """
        Calculates the Laplacian.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the Laplacians at each point.
        """
        return self._hessian_functional(x=x, matrix_functional=np.trace)

    def hessian_determinant(self, x: np.ndarray) -> np.ndarray:
        """
        Calculates the determinant of the Hessian matrix.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the Hessian determinants at each point.
        """
        return self._hessian_functional(x=x, matrix_functional=np.linalg.det)

    def jacobian_norm(self, x: np.ndarray) -> np.ndarray:
        """
        Gradient 2-norm.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the norm of the gradient at each point.
        """
        return np.linalg.norm(self.jacobian(x), axis=1)

    def integral(self, ranges: List[Tuple[float, float]]) -> float:
        """
        Calculates the integral of the pdf over a rectangular domain.

        :param ranges: List of 2-tuples defining the rectangular integration
         domain.
        :return: The total probability mass enclosed in the integration domain.
        """

        def pdf(*x):
            return float(self(np.array([list(x)])))

        return nquad(func=pdf, ranges=ranges)[0]

    @abstractmethod
    def _pdf(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the pdf at the input points, without checking input
        dimensions.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: An array with as many elements as input points corresponding
         to the pdf values of each point.
        """
        pass

    @abstractmethod
    def _jacobian(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the Jacobian of the pdf at the input points, without checking
        input dimensions.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: A matrix with as many rows as input points, where each column
         corresponds to one of the partial derivatives.
        """
        pass

    @abstractmethod
    def _hessian(self, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the Hessian of the pdf at the input points, without checking
        input dimensions.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: A matrix with as many rows as input points, where each column
         corresponds to one of the second-order partial derivatives.
        """
        pass

    @abstractmethod
    def _derivative(self, r: int, x: np.ndarray) -> np.ndarray:
        """
        Evaluates the r-th order derivative of the pdf at the input points,
        without checking input dimensions.

        :param r: Order of the derivatives.
        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :return: A matrix with as many rows as input points, where each column
         corresponds to one of the r-th order partial derivatives.
        """
        pass

    def _is_valid_input(self, x: np.ndarray) -> bool:
        """
        Determines whether a multidimensional array is a valid pdf input,
        representing points with a suitable dimension.

        :param x: Input tensor to be tested.
        :return: true if and only if the input is valid.
        """
        return len(x.shape) == 2 and x.shape[1] == self.dimension

    def _check_input(self, x: np.ndarray):
        """
        Raises an exception if the input tensor is not a valid argument for the
        pdf, representing points with a suitable dimension.

        :param x: Input tensor to be tested.
        """
        if not self._is_valid_input(x):
            raise ValueError("Incorrect input dimensions")

    def _unvectorize(self, x: np.ndarray) -> np.ndarray:
        """
        Reverts the vectorization for (d x d)-element vectors, transforming
        them back to d x d matrices.

        :param x: Vector to be back-transformed to a matrix.
        :return: Reshaped matrix with elements from the input vector.
        """
        return x.reshape((self.dimension, self.dimension))

    @staticmethod
    def _kronecker_jacobian(jacobian: np.ndarray):
        """
        Calculates the matrix based on first-order partial derivatives that
        is used to calculate the shape operator.

        :param jacobian: Jacobian with partial derivatives.
        :return: The auxiliary matrix.
        """
        identity: Final = np.eye(len(jacobian))
        inner_prod: Final = 1 + np.dot(jacobian, jacobian)
        outer_prod: Final = np.outer(jacobian, jacobian)
        return np.kron(identity, identity - outer_prod / inner_prod) / np.sqrt(
            inner_prod
        )

    def _matrix_functional(
        self,
        x: np.ndarray,
        matrix_feature: Callable[[np.ndarray], np.ndarray],
        matrix_functional: Callable[[np.ndarray], float],
    ) -> np.ndarray:
        """
        Applies a functional to a matrix-valued pdf features.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :param matrix_feature: Matrix-valued pdf feature.
        :param matrix_functional: Functional to apply to each matrix feature.
        :return: An array with as many elements as input rows comprising the
         values from applying the functional to the matrix-valued feature at
         the corresponding point.
        """
        return np.apply_along_axis(
            func1d=lambda vec_matrix: matrix_functional(
                self._unvectorize(vec_matrix)
            ),
            axis=1,
            arr=matrix_feature(x),
        )

    def _hessian_functional(
        self, x: np.ndarray, matrix_functional: Callable[[np.ndarray], float]
    ) -> np.ndarray:
        """
        Applies a functional to the Hessian matrix.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :param matrix_functional: Functional to apply to each matrix feature.
        :return: An array with as many elements as input rows comprising the
         values from applying the functional to the Hessian matrix.
        """
        return self._matrix_functional(
            x=x,
            matrix_feature=self.hessian,
            matrix_functional=matrix_functional,
        )

    def _shape_functional(
        self, x: np.ndarray, matrix_functional: Callable[[np.ndarray], float]
    ) -> np.ndarray:
        """
        Applies a functional to the shape operator matrix.

        :param x: Input points. Each row represents a different point, being
         each column a point component.
        :param matrix_functional: Functional to apply to each matrix feature.
        :return: An array with as many elements as input rows comprising the
         values from applying the functional to the shape operator matrix.
        """
        return self._matrix_functional(
            x=x, matrix_feature=self.shape, matrix_functional=matrix_functional
        )


class Mixture(PDF):
    """
    A pdf built from a weighted average of pdfs.
    """

    def __init__(self, pdfs: List[PDF], weights: List[float]):
        """
        Initializes the mixture with its components.

        :param pdfs: Component pdfs.
        :param weights: Weights for each pdf component.
        """
        if not all(np.array(weights) > 0):
            raise ValueError("There are non-positive weights")

        if np.sum(np.array(weights)) != 1.0:
            raise ValueError("Weights do not sum up to one")

        if len(pdfs) != len(weights):
            raise ValueError("Different number of pdfs and weights")

        if not all([pdf.dimension == pdfs[0].dimension for pdf in pdfs]):
            raise ValueError("Densities with different dimensions")

        self._pdfs: Final = pdfs
        self._weights: Final = weights

    @property
    def dimension(self) -> int:
        return self._pdfs[0].dimension

    def sample(
        self, size: int, rand: np.random.RandomState = None
    ) -> np.ndarray:

        if rand is None:
            rand = np.random

        mixtures = rand.choice(
            a=range(len(self._pdfs)), size=size, p=self._weights
        )

        components, sizes = np.unique(mixtures, return_counts=True)

        return np.vstack(
            [
                self._pdfs[component].sample(size=_size, rand=rand)
                for component, _size in zip(components, sizes)
            ]
        )

    def _pdf(self, x: np.ndarray) -> np.ndarray:
        return self._weighted_sum(x=x, feature=lambda pdf, _x: pdf(_x))

    def _jacobian(self, x: np.ndarray) -> np.ndarray:
        return self._weighted_sum(x=x, feature=lambda pdf, _x: pdf.jacobian(_x))

    def _hessian(self, x: np.ndarray) -> np.ndarray:
        return self._weighted_sum(x=x, feature=lambda pdf, _x: pdf.hessian(_x))

    def _derivative(self, r: int, x: np.ndarray) -> np.ndarray:
        return self._weighted_sum(
            x=x, feature=lambda pdf, _x: pdf.derivative(r=r, x=_x)
        )

    def _weighted_sum(
        self, x: np.ndarray, feature: Callable[[PDF, np.ndarray], np.ndarray]
    ):
        """
        Calculates a feature by averaging the corresponding features from the
        components.

        :param x: Points where to calculate the feature.
        :param feature: Feature to be calculated.
        :return: The feature values at the input points.
        """
        return np.add.reduce(
            [
                weight * feature(pdf, x)
                for pdf, weight in zip(self._pdfs, self._weights)
            ],
        )
