"""
Gaussian densities and mixtures.
"""
from typing import Final

import numpy as np
from scipy.linalg import det, inv, sqrtm
from scipy.stats import norm

from kdfeatures.density import PDF
from kdfeatures.util.hermite import hermite_minimal_polynomial
from kdfeatures.util.matrix import is_inner_product


def gaussian_pdf(x: np.ndarray) -> np.ndarray:
    """
    Evaluates the Gaussian pdf. Each row in the input corresponds to a
    different point to be evaluated. Each column represents a different
    point component.

    :param x: Points to evaluate the pdf at.
    :return: A vector with as many elements as input rows indicating the value
     of the pdf at the respective point.
    """
    if len(x.shape) != 2:
        raise ValueError("Input tensor must have dimension 2")

    return np.prod(norm.pdf(x), axis=1)


def gaussian_jacobian(x: np.ndarray, pdf: np.ndarray = None) -> np.ndarray:
    """
    Evaluates the Gaussian gradient. Each row in the input corresponds to a
    different point to be evaluated. Each column represents a different
    point component.

    :param x: Points to evaluate the Jacobian at.
    :param pdf: pdf value, if it has already been calculated.
    :return: A matrix with as many rows as input, where each column corresponds
     to a different partial derivative.
    """
    if len(x.shape) != 2:
        raise ValueError("Input tensor must have dimension 2")

    if pdf is None:
        pdf = gaussian_pdf(x)

    return -pdf.reshape(-1, 1) * x


def _kron_square(x: np.ndarray) -> np.ndarray:
    """
    Performs the row-wise Kronecker product of a 2-tensor with itself.

    :param x: Input 2-tensor.
    :return: The row-wise Kronecker product of the input 2-tensor with itself.
    """
    if x.shape[1] == 1:
        return (x[:, 0] ** 2).reshape(-1, 1)
    elif x.shape[1] == 2:
        x01: Final = x[:, 0] * x[:, 1]
        return np.column_stack([x[:, 0] ** 2, x01, x01, x[:, 1] ** 2])
    elif x.shape[1] == 3:
        x01: Final = x[:, 0] * x[:, 1]
        x12: Final = x[:, 1] * x[:, 2]
        x02: Final = x[:, 0] * x[:, 2]
        return np.column_stack(
            [
                x[:, 0] ** 2,
                x01,
                x02,
                x01,
                x[:, 1] ** 2,
                x12,
                x02,
                x12,
                x[:, 2] ** 2,
            ]
        )
    else:
        return np.apply_along_axis(
            func1d=lambda _x: np.kron(_x, _x), axis=1, arr=x
        )


def gaussian_hessian(x: np.ndarray, pdf: np.ndarray = None) -> np.ndarray:
    """
    Evaluates the Gaussian Hessian. Each row in the input corresponds to a
    different point to be evaluated. Each column represents a different
    point component.

    :param x: Points to evaluate the Hessian at.
    :param pdf: pdf value, if it has already been calculated.
    :return: A matrix with as many rows as input, where each column corresponds
     to a different second-order partial derivative.
    """
    if len(x.shape) != 2:
        raise ValueError("Input tensor must have dimension 2")

    if pdf is None:
        pdf = gaussian_pdf(x)

    return pdf.reshape(-1, 1) * (_kron_square(x) - np.eye(x.shape[1]).flatten())


class Gaussian(PDF):
    """
    Gaussian probability density function.
    """

    def __init__(
        self,
        dimension: int,
        mean: np.ndarray = None,
        covariance: np.ndarray = None,
    ):
        """
        Initializes the Gaussian pdf.

        :param dimension: Dimension of the sampling space.
        :param mean: Means vector.
        :param covariance: Covariance matrix.
        """
        # Check distribution parameters.

        Gaussian._check_parameters(dimension, mean, covariance)

        # Internals.

        self._dimension: Final = dimension

        self._mean: Final = mean
        self._covariance: Final = covariance

        self._h: Final = (
            inv(sqrtm(covariance)) if covariance is not None else None
        )
        self._det_h: Final = det(self._h) if self._h is not None else None

        self._h2: Final = (
            np.kron(self._h, self._h) if self._h is not None else None
        )

        if self._det_h is not None:
            assert self._det_h > 0

    @property
    def dimension(self) -> int:
        return self._dimension

    def sample(
        self, size: int, rand: np.random.RandomState = None
    ) -> np.ndarray:
        if rand is None:
            rand = np.random

        return rand.multivariate_normal(
            mean=self.mean, cov=self.covariance, size=size
        )

    @property
    def mean(self) -> np.ndarray:
        """
        Returns the mean vector.

        :return: The mean vector.
        """
        return (
            self._mean if self._mean is not None else np.zeros(self.dimension)
        )

    @property
    def covariance(self) -> np.ndarray:
        """
        Returns the covariance matrix.

        :return: The covariance matrix.
        """
        return (
            self._covariance
            if self._covariance is not None
            else np.eye(self.dimension)
        )

    def _pdf(self, x: np.ndarray) -> np.ndarray:
        pdf = gaussian_pdf(self._transformed_x(x))

        if self._det_h is not None:
            pdf *= self._det_h

        return pdf

    def _jacobian(self, x: np.ndarray) -> np.ndarray:
        jac = gaussian_jacobian(self._transformed_x(x))

        if self._h is not None:
            jac = np.dot(jac, self._h)

        if self._det_h is not None:
            jac *= self._det_h

        return jac

    def _hessian(self, x: np.ndarray) -> np.ndarray:
        hess = gaussian_hessian(self._transformed_x(x))

        if self._h2 is not None:
            hess = np.dot(hess, self._h2)

        if self._det_h is not None:
            hess *= self._det_h

        return hess

    def _derivative(self, r: int, x: np.ndarray) -> np.ndarray:
        if r < 0:
            raise ValueError("Wrong density derivative order")

        h = hermite_minimal_polynomial(
            x=x - self.mean, v=inv(self.covariance), order=r
        )

        return ((-1) ** r) * h.enlarged_vector() * self._pdf(x).reshape((-1, 1))

    def _transformed_x(self, x: np.ndarray) -> np.ndarray:
        """
        Subtracts mean and applies covariance scaling to pdf input vector.

        :param x: Input vectors to be transformed.
        :return: The input vectors centred and scaled.
        """
        if self._mean is not None or self._h is not None:
            transformed_x = np.copy(x)
        else:
            transformed_x = x

        if self._mean is not None:
            transformed_x -= self._mean

        if self._h is not None:
            transformed_x = np.dot(transformed_x, self._h)

        return transformed_x

    @staticmethod
    def _check_parameters(
        dimension: int, mean: np.ndarray, covariance: np.ndarray
    ):
        """
        Checks Gaussian distribution parameters, raising an exception if
        an error is found.

        :param dimension: Dimension of the multivariate distribution.
        :param mean: Mean vector.
        :param covariance: Covariance matrix.
        """

        if dimension < 1:
            raise ValueError("Dimension should be at least 1")

        if mean is not None:
            if mean.shape != (dimension,):
                raise ValueError("Incorrect mean dimension")

        if covariance is not None:
            if not is_inner_product(covariance):
                raise ValueError("Matrix is not a covariance matrix")

            if covariance.shape[0] != dimension:
                raise ValueError("Incorrect covariance dimensions")


def univariate_gaussian(mean: float = None, sigma: float = None):
    """
    Builds a univariate Gaussian pdf.

    :param mean: Mean of the pdf. Defaults to zero.
    :param sigma: Standard deviation of the pdf. Defaults to one.
    :return: A univariate Gaussian pdf with the given parameters.
    """
    return Gaussian(
        1,
        mean=np.array([mean]) if mean is not None else None,
        covariance=np.array([[sigma**2]]) if sigma is not None else None,
    )


def bivariate_gaussian(
    mean_x: float = None,
    mean_y: float = None,
    sigma_x: float = None,
    sigma_y: float = None,
    corr: float = None,
) -> Gaussian:
    """
    Builds a bivariate Gaussian pdf.

    :param mean_x: Mean of the x variable.
    :param mean_y: Mean of the y variable.
    :param sigma_x: Standard deviation of the x variable.
    :param sigma_y: Standard deviation of the y variable.
    :param corr: Correlation between the x and y variables.
    :return: The configured bivariate Gaussian pdf.
    """

    if mean_x is not None or mean_y is not None:
        if mean_x is None:
            mean_x = 0.0

        if mean_y is None:
            mean_y = 0.0

        mean = np.array([mean_x, mean_y])
    else:
        mean = None

    if sigma_x is not None or sigma_y is not None or corr is not None:
        if sigma_x is None:
            sigma_x = 1.0

        if sigma_y is None:
            sigma_y = 1.0

        if corr is None:
            corr = 0.0

        cov = corr * sigma_x * sigma_y

        covariance = np.array([[sigma_x**2, cov], [cov, sigma_y**2]])
    else:
        covariance = None

    return Gaussian(2, mean=mean, covariance=covariance)


def trivariate_gaussian(
    mean_x: float = None,
    mean_y: float = None,
    mean_z: float = None,
    sigma_x: float = None,
    sigma_y: float = None,
    sigma_z: float = None,
    corr_xy: float = None,
    corr_yz: float = None,
    corr_xz: float = None,
) -> Gaussian:
    """
    Builds a trivariate Gaussian pdf.

    :param mean_x: Mean of the x variable.
    :param mean_y: Mean of the y variable.
    :param mean_z: Mean of the z variable.
    :param sigma_x: Standard deviation of the x variable.
    :param sigma_y: Standard deviation of the y variable.
    :param sigma_z: Standard deviation of the z variable.
    :param corr_xy: Correlation between the x and y variables.
    :param corr_yz: Correlation between the y and z variables.
    :param corr_xz: Correlation between the x and z variables.
    :return: The configured bivariate Gaussian pdf.
    """

    if mean_x is not None or mean_y is not None or mean_z is not None:
        if mean_x is None:
            mean_x = 0.0

        if mean_y is None:
            mean_y = 0.0

        if mean_z is None:
            mean_z = 0.0

        mean = np.array([mean_x, mean_y, mean_z])
    else:
        mean = None

    if (
        sigma_x is not None
        or sigma_y is not None
        or sigma_z is not None
        or corr_xy is not None
        or corr_yz is not None
        or corr_xz is not None
    ):
        if sigma_x is None:
            sigma_x = 1.0

        if sigma_y is None:
            sigma_y = 1.0

        if sigma_z is None:
            sigma_z = 1.0

        if corr_xy is None:
            corr_xy = 0.0

        if corr_yz is None:
            corr_yz = 0.0

        if corr_xz is None:
            corr_xz = 0.0

        cov_xy: Final = corr_xy * sigma_x * sigma_y
        cov_yz: Final = corr_yz * sigma_y * sigma_z
        cov_xz: Final = corr_xz * sigma_x * sigma_z

        covariance = np.array(
            [
                [sigma_x**2, cov_xy, cov_xz],
                [cov_xy, sigma_y**2, cov_yz],
                [cov_xz, cov_yz, sigma_z**2],
            ]
        )
    else:
        covariance = None

    return Gaussian(3, mean=mean, covariance=covariance)
