"""
Kernel density estimation.
"""
from typing import Callable, Final, List, Tuple, Union

import numpy as np

from kdfeatures.density import Gaussian, PDF
from kdfeatures.density.bootstrap import bootstrap_statistic
from kdfeatures.util.grid import rectangular_grid
from kdfeatures.util.matrix import (
    cartesian_differences,
    is_inner_product,
    is_matrix,
)
from kdfeatures.util.quant import tail_value_at_risk


class KDE(PDF):
    """
    Kernel density estimator.
    """

    def __init__(
        self,
        sample: np.ndarray,
        bandwidth: np.ndarray,
    ):
        """
        Initializes the KDE.

        :param sample: Sample that builds the KDE.
        :param bandwidth: Bandwidths matrix.
        """
        if not is_matrix(sample):
            raise ValueError("Sample is not a matrix")

        if not is_inner_product(bandwidth):
            raise ValueError("Bandwidth is not a covariance matrix")

        if bandwidth.shape[0] != sample.shape[1]:
            raise ValueError("Incorrect bandwidth dimensions")

        self._sample: Final = sample
        self._bandwidth: Final = bandwidth

        # Internals.

        self._gaussian: Final = Gaussian(
            dimension=self.dimension, covariance=bandwidth
        )

    @property
    def dimension(self) -> int:
        return self._bandwidth.shape[0]

    @property
    def underlying_sample(self) -> np.ndarray:
        """
        Returns the underlying sample giving rise to the KDE.

        :return: The underlying sample giving rise to the KDE.
        """
        return self._sample

    @property
    def bandwidth(self) -> np.ndarray:
        """
        Returns the KDE bandwidth matrix.

        :return: The KDE bandwidth matrix.
        """
        return self._bandwidth

    def quantile(self, probability: float) -> float:
        """
        Given a probability value p, between 0 and 1, calculates a pdf value q
        such that a proportion p of the underlying sample observations have
        a pdf value not smaller than q.

        :param probability: Total mass to enclose in the upper level set.
        :return: PDF threshold value that determines a highest density region
         enclosing the given probability mass.
        """
        return np.quantile(a=self(self.underlying_sample), q=1 - probability)

    @property
    def sample_size(self) -> int:
        """
        Returns the number of observations in the underlying sample.

        :return: The number of observations in the underlying sample.
        """
        return len(self._sample)

    def sample(
        self, size: int, rand: np.random.RandomState = None
    ) -> np.ndarray:

        centers = rand.choice(a=range(len(self._sample)), size=size)
        normal_sample = self._gaussian.sample(size=size, rand=rand)

        return np.vstack(
            [self._sample[centers[i]] + normal_sample[i] for i in range(size)]
        )

    def _pdf(self, x: np.ndarray) -> np.ndarray:
        return self._average(x=x, f=lambda _x: self._gaussian(_x))

    def _jacobian(self, x: np.ndarray) -> np.ndarray:
        return self._average(x=x, f=lambda _x: self._gaussian.jacobian(_x))

    def _hessian(self, x: np.ndarray) -> np.ndarray:
        return self._average(x=x, f=lambda _x: self._gaussian.hessian(_x))

    def _derivative(self, r: int, x: np.ndarray) -> np.ndarray:
        return self._average(
            x=x, f=lambda _x: self._gaussian.derivative(r=r, x=_x)
        )

    def _average(
        self, x: np.ndarray, f: Callable[[np.ndarray], np.ndarray]
    ) -> np.ndarray:
        """
        Performs an average of a kernel functional over the whole sample.

        :param x: Points at which to evaluate the kernel functional.
        :param f: Kernel functional to evaluate.
        :return: The average of the kernel functional at the input points.
        """

        f_batch = f(cartesian_differences(x=x, y=self._sample))

        if len(f_batch.shape) == 1:
            f_batch = f_batch.reshape((x.shape[0], -1))
        else:
            f_batch = f_batch.reshape((x.shape[0], -1, f_batch.shape[1]))

        return np.sum(f_batch, axis=1) / len(self._sample)

    def _bootstrap_kde(self, bootstrap_sample: np.ndarray):
        """
        Instantiates a KDE based on a bootstrap sample with the same bandwidth
        matrix.

        :param bootstrap_sample: Bootstrap sample to build the KDE from.
        :return: A KDE with the same bandwidth as self and based on the input
         bootstrap sample.
        """
        return KDE(sample=bootstrap_sample, bandwidth=self.bandwidth)

    def _bootstrap_error_laplacian(
        self,
        bootstrap_sample: np.ndarray,
        *grid_configs: Tuple[float, float, int],
    ) -> float:
        """
        Calculates the Laplacian bootstrap absolute error over a certain
        bootstrap sample.

        :param bootstrap_sample: Bootstrap sample to calculate the error.
        :param grid_configs: Rectangular grid where to compare the original KDE
         with the bootstrap KDE.
        :return: The supremum bootstrap error for the Laplacian.
        """
        if len(grid_configs) != self.dimension:
            raise ValueError("Wrong dimensions for rectangular grid")

        # Build bootstrap KDE.

        bootstrap_kde: Final = self._bootstrap_kde(bootstrap_sample)

        # Evaluate the Laplacian bootstrap error over a rectangular grid.

        grid, _ = rectangular_grid(*grid_configs)

        laplacian_kde: Final = self.laplacian(grid)
        laplacian_bootstrap_kde: Final = bootstrap_kde.laplacian(grid)

        assert laplacian_kde.shape == laplacian_bootstrap_kde.shape

        return float(np.max(np.abs(laplacian_kde - laplacian_bootstrap_kde)))

    def _bootstrap_error_hessian(
        self,
        bootstrap_sample: np.ndarray,
        *grid_configs: Tuple[float, float, int],
    ) -> np.ndarray:
        """
        Calculates the supremum absolute error over a certain bootstrap sample
        for all second-order partial derivatives of the KDE.

        :param bootstrap_sample: Bootstrap sample to calculate the error.
        :param grid_configs: Rectangular grid where to compare the original KDE
         with the bootstrap KDE.
        :return: The supremum bootstrap error for the second-order derivatives.
        """
        if len(grid_configs) != self.dimension:
            raise ValueError("Wrong dimensions for rectangular grid")

        # Build bootstrap KDE.

        bootstrap_kde: Final = self._bootstrap_kde(bootstrap_sample)

        # Evaluate the Hessian bootstrap error over a rectangular grid.

        grid, _ = rectangular_grid(*grid_configs)

        hessian_kde: Final = self.hessian(grid)
        hessian_bootstrap_kde: Final = bootstrap_kde.hessian(grid)
        assert hessian_kde.shape == hessian_bootstrap_kde.shape

        supnorms: Final = np.max(
            np.abs(hessian_kde - hessian_bootstrap_kde), axis=0
        )

        assert supnorms.shape == (self.dimension**2,)

        return supnorms

    def _bootstrap_statistic(
        self,
        statistic: Callable[[np.ndarray], Union[float, np.ndarray]],
        num_samples: int,
        resample_size: int = None,
        rand: np.random.RandomState = None,
    ) -> np.ndarray:
        """
        Calculates a certain statistic over bootstrap samples derived from
        the underlying KDE sample.

        :param statistic: Statistic to calculate.
        :param num_samples: Number of bootstrap samples to generate.
        :param resample_size: Number of observations per bootstrap sample.
        :param rand: Random generator.
        :return: Either a one-dimensional array or a bidimensional array, where
         each row corresponds to a vector of statistic from a bootstrap sample.
        """
        s: Final = bootstrap_statistic(
            statistic=statistic,
            sample=self.underlying_sample,
            num_samples=num_samples,
            resample_size=resample_size,
            rand=rand,
        )

        if isinstance(s[0], float):
            return np.array(s)
        else:
            return np.vstack(s)

    def laplacian_bootstrap_margin(
        self,
        num_samples: int,
        confidence: float,
        resample_size: int = None,
        rand: np.random.RandomState = None,
        grid_configs: List[Tuple[float, float, int]] = None,
    ) -> float:
        """
        Calculates the required bound for building confidence sets enclosing
        Laplacian bumps.

        :param num_samples: Number of bootstrap samples.
        :param confidence: Confidence level in (0, 1).
        :param resample_size: Size of bootstrap samples.
        :param rand: Random generator.
        :param grid_configs: Rectangular grid where to compare the original KDE
         with the bootstrap KDE.
        :return: The margin needed to build confidence sets for Laplacian bumps.
        """
        if not (0 < confidence < 1):
            raise ValueError("Confidence is not in (0, 1)")

        if grid_configs is None:
            grid_configs = [(-10.0, 10.0, 101)] * self.dimension

        errors: Final = self._bootstrap_statistic(
            statistic=lambda s: self._bootstrap_error_laplacian(
                s, *grid_configs
            ),
            num_samples=num_samples,
            resample_size=resample_size,
            rand=rand,
        )

        return float(np.quantile(a=errors, q=confidence))

    def eigenvalue_bootstrap_margin(
        self,
        num_samples: int,
        confidence: float,
        resample_size: int = None,
        rand: np.random.RandomState = None,
        grid_configs: List[Tuple[float, float, int]] = None,
    ) -> float:
        """
        Calculates the required bound for building confidence sets enclosing
        concave bumps.

        :param num_samples: Number of bootstrap samples.
        :param confidence: Confidence level in (0, 1).
        :param resample_size: Size of bootstrap samples.
        :param rand: Random generator.
        :param grid_configs: Rectangular grid where to compare the original KDE
         with the bootstrap KDE.
        :return: The margin needed to build confidence sets for concave bumps.
        """

        if not (0 < confidence < 1):
            raise ValueError("Confidence is not in (0, 1)")

        if grid_configs is None:
            grid_configs = [(-10.0, 10.0, 101)] * self.dimension

        derivatives_errors = self._bootstrap_statistic(
            statistic=lambda s: self._bootstrap_error_hessian(s, *grid_configs),
            num_samples=num_samples,
            resample_size=resample_size,
            rand=rand,
        )

        assert derivatives_errors.shape == (num_samples, self.dimension**2)

        derivatives_errors = np.split(derivatives_errors.T, self.dimension**2)
        assert len(derivatives_errors) == self.dimension**2

        bound = 0.0

        for derivative_errors in derivatives_errors:
            derivative_errors = np.squeeze(derivative_errors)
            assert len(derivative_errors) == num_samples

            bound += tail_value_at_risk(derivative_errors, p=confidence)

        return bound


def univariate_kde(sample: np.ndarray, h: float) -> KDE:
    """
    Instantiates a univariate KDE.

    :param sample: Sample to build the KDE.
    :param h: Bandwith parameter.
    :return: The configured univariate KDE.
    """
    return KDE(
        sample=np.expand_dims(sample, axis=1)
        if len(sample.shape) != 2
        else sample,
        bandwidth=np.array([[h]]),
    )
