"""
Probability density function features.
"""
import warnings
from abc import ABC, abstractmethod
from typing import Final

import numpy as np
from matplotlib import cm, colors
from matplotlib.cm import ScalarMappable
from scipy.optimize import fsolve

from kdfeatures.density import PDF


class PDFFeature(ABC):
    """
    An abstract real-valued feature that can be calculated from a pdf over a
    rectangular grid and graphically represented with colors.
    """

    def __init__(self):
        """
        Creates the feature.
        """
        self._pdf = None
        self._values = None

    def __call__(self, pdf: PDF, x: np.ndarray):
        """
        Initializes the feature, calculating the values over the grid.

        :param pdf: Density function from which to extract the feature values.
        :param x: Points in which to evaluate the pdf feature.
        """
        self._pdf = pdf
        self._values = self.calculate_values(x=x)

    @property
    def values(self) -> np.ndarray:
        """
        The calculated feature values over the input grid.

        :return: The calculated feature values as a matrix with the same shape
         as the horizontal and vertical grids.
        """
        return self._values

    @abstractmethod
    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        """
        Calculates the feature values over and array of points.

        :param x: Array of points over which the feature must be calculated.
        :return: An array with as many elements as in the input with the
         corresponding calculated values from the pdf feature.
        """
        pass

    @abstractmethod
    def coloring(self) -> ScalarMappable:
        """
        Builds the color mapping for each feature value.

        :return: The colors corresponding to the calculated feature values as
         a matrix.
        """
        pass


class PDFUnsignedFeature(PDFFeature):
    """
    A pdf feature taking only non-negative values.
    """

    def coloring(self) -> ScalarMappable:
        mappable = ScalarMappable(cmap=cm.hot)
        mappable.set_array(self.values)
        return mappable


class PDFSignedFeature(PDFFeature):
    """
    A pdf feature taking either zero, negative or positive values.
    """

    @property
    @abstractmethod
    def reference_sign(self) -> float:
        """
        Indicates the sign that takes the feature in the interior of a bump.

        :return: The sign, 1 or -1, that takes the feature in the interior of a
         bump.
        """
        pass

    def coloring(self) -> ScalarMappable:
        mappable = ScalarMappable(
            cmap=cm.seismic,
            norm=colors.TwoSlopeNorm(
                vcenter=0.0,
            ),
        )
        mappable.set_array(self.values)

        return mappable


class PDFUnsignedHeight(PDFUnsignedFeature):
    """
    The height of the pdf.
    """

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf(x)


class PDFSignedHeight(PDFSignedFeature):
    """
    The height of the pdf minus a certain threshold.
    """

    def __init__(self, threshold: float):
        """
        Creates the feature.

        :param threshold: Value that determines the feature boundary.
        """
        super().__init__()

        self._threshold: Final = threshold

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf(x) - self._threshold

    @property
    def reference_sign(self):
        return 1.0


class PDFGradientNorm(PDFUnsignedFeature):
    """
    Norm of the pdf gradient.
    """

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf.jacobian_norm(x)


class PDFLargestHessianEigenvalue(PDFSignedFeature):
    """
    The largest eigenvalue of the pdf Hessian matrix.
    """

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf.largest_hessian_eigenvalue(x)

    @property
    def reference_sign(self):
        return -1.0


class PDFSmallestHessianEigenvalue(PDFSignedFeature):
    """
    The smallest eigenvalue of the pdf Hessian matrix.
    """

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf.smallest_hessian_eigenvalue(x)

    @property
    def reference_sign(self):
        return 1.0


class PDFMeanCurvature(PDFSignedFeature):
    """
    The mean curvature.
    """

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf.mean_curvature(x)

    @property
    def reference_sign(self):
        return -1.0


class PDFGaussianCurvature(PDFSignedFeature):
    """
    The Gaussian curvature.
    """

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf.gaussian_curvature(x)

    @property
    def reference_sign(self):
        return 1.0


class PDFLaplacian(PDFSignedFeature):
    """
    The Laplacian.
    """

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf.laplacian(x)

    @property
    def reference_sign(self):
        return -1.0


class PDFHessianDeterminant(PDFSignedFeature):
    """
    The Laplacian.
    """

    def calculate_values(self, x: np.ndarray) -> np.ndarray:
        return self._pdf.hessian_determinant(x)

    @property
    def reference_sign(self):
        return 1.0


def feature_roots(
    feature: PDFSignedFeature, xmin: float, xmax: float, fsolve_x0_size: int
):
    """
    Finds the roots of a signed univariate pdf feature.

    :param feature: Feature to extract roots from.
    :param xmin: Minimum value in the x axis.
    :param xmax: Maximum value in the x axis.
    :param fsolve_x0_size: Number of initial guesses to fin all roots of the
     contour feature function.
    :return: An array with the roots.
    """

    @np.vectorize
    def f(z: float):
        return float(feature.calculate_values(np.array([[z]])))

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        return [
            root
            for root in np.unique(
                fsolve(
                    func=f,
                    x0=np.linspace(start=xmin, stop=xmax, num=fsolve_x0_size),
                )
            )
            if xmin <= root <= xmax
        ]
