"""
Confidence set plotting.
"""

import matplotlib.pyplot as plt
import numpy as np

from kdfeatures.density import KDE, PDF
from kdfeatures.plot import (
    PDFLaplacian,
    PDFLargestHessianEigenvalue,
    PDFSignedFeature,
)
from kdfeatures.util.grid import rectangular_grid


def plot_confidence_sets_2d(
    kde: KDE,
    confidence: float,
    num_bootstrap_samples: int,
    feature: PDFSignedFeature,
    xmin: float,
    xmax: float,
    numx: int,
    ymin: float,
    ymax: float,
    numy: int,
    pdf: PDF = None,
    scatter_size: float = None,
    contour_linewidth: float = 3.0,
    contour_opacity: float = 0.5,
    rand: np.random.RandomState = None,
):
    """
    Plots confidence sets for bidimensional curvature bumps.

    :param kde: KDE from which to extract bumps.
    :param confidence: Desired probability of the true bump being comprised
     between the confidence sets.
    :param num_bootstrap_samples: Number of bootstrap samples.
    :param feature: Curvature feature corresponding to the requested bump.
    :param xmin: Minimum value in the x axis.
    :param xmax: Maximum value in the x axis.
    :param numx: Number of points in the x axis.
    :param ymin: Minimum value in the y axis.
    :param ymax: Maximum value in the y axis.
    :param numy: Number of points in the y axis.
    :param pdf: Original PDF. Provided to draw the true bump boundary.
    :param scatter_size: Size of observation markers.
    :param contour_linewidth: Width of the true bump boundary.
    :param contour_opacity: A number between 0 (transparent) and 1 (opaque)
     for confidence sets.
    :param rand: Random generator.
    :return: The generated axes.
    """
    if isinstance(feature, PDFLaplacian):
        bound = kde.laplacian_bootstrap_margin(
            num_samples=num_bootstrap_samples,
            confidence=confidence,
            rand=rand,
            grid_configs=[(xmin, xmax, numx), (ymin, ymax, numy)],
        )
    elif isinstance(feature, PDFLargestHessianEigenvalue):
        bound = kde.eigenvalue_bootstrap_margin(
            num_samples=num_bootstrap_samples,
            confidence=confidence,
            rand=rand,
            grid_configs=[(xmin, xmax, numx), (ymin, ymax, numy)],
        )
    else:
        raise NotImplemented(
            "Confidence intervals not implemented for that feature"
        )

    levels = [
        -np.Inf,
        -bound,
        0.0,
        bound,
        np.Inf,
    ]

    #
    # Scatter plot.
    #

    _, ax = plt.subplots()

    ax.scatter(
        x=kde.underlying_sample[:, 0],
        y=kde.underlying_sample[:, 1],
        color="w",
        s=scatter_size,
        edgecolors="k",
    )

    #
    # Bump boundary plot.
    #

    # Rectangular grid.

    xy, (x, y) = rectangular_grid((xmin, xmax, numx), (ymin, ymax, numy))

    # Plot bump boundary.

    if pdf is not None:
        feature(pdf=pdf, x=xy)

        ax.contour(
            x,
            y,
            feature.values.reshape(x.shape),
            levels=[0.0],
            colors=["crimson"],
            linewidths=contour_linewidth,
        )

    #
    # Confidence sets plot.
    #

    feature(pdf=kde, x=xy)

    ax.contourf(
        x,
        y,
        feature.values.reshape(x.shape),
        levels=levels,
        colors=["navy", "lightseagreen", "aquamarine", "sandybrown"],
        alpha=contour_opacity,
    )

    ax.set_xlim(left=xmin, right=xmax)
    ax.set_ylim(bottom=ymin, top=ymax)

    return ax
