"""
Random sample plotting.
"""
from typing import List

import numpy as np
from matplotlib import pyplot as plt
from skimage.measure import marching_cubes

from kdfeatures.density import PDF
from kdfeatures.plot import PDFFeature, PDFSignedFeature
from kdfeatures.plot.feature import feature_roots
from kdfeatures.util.grid import rectangular_grid


def plot_sample_1d(
    pdf: PDF,
    scatter_feature: PDFFeature = None,
    contour_feature: PDFSignedFeature = None,
    xmin: float = None,
    xmax: float = None,
    numx: int = 101,
    sample: np.ndarray = None,
    size: int = None,
    rand: np.random.RandomState = None,
    scatter_size: float = None,
    fsolve_x0_size: int = 21,
    colorbar_shrink: float = 1.0,
    ax: plt.Axes = None,
):
    """
    Plots a random sample from a given univariate pdf along with the zero-level
    contour of one pdf feature.

    :param pdf: Univariate density to plot.
    :param scatter_feature: Feature to color scatter dots.
    :param contour_feature: pdf feature to extract the zero-level set.
    :param xmin: Minimum value in the x axis.
    :param xmax: Maximum value in the x axis.
    :param numx: Number of points in the x axis.
    :param sample: Sample to plot.
    :param size: Number of random observations to draw from the pdf.
    :param rand: Random generator.
    :param scatter_size: Size of observation markers.
    :param fsolve_x0_size: Number of initial guesses to fin all roots of the
     contour feature function.
    :param colorbar_shrink: Shrink factor for color bar.
    :param ax: Axes to be utilized when plotting.
    :return: The generated axes.
    """

    if pdf.dimension != 1:
        raise ValueError("Density is not 1-dimensional")

    if sample is None:
        if size is None:
            raise ValueError(
                "Either a sample or a sample size must be provided"
            )

        sample = pdf.sample(size=size, rand=rand)

    # Bounds.

    if xmin is None:
        xmin = np.amin(sample[:, 0])

    if xmax is None:
        xmax = np.amax(sample[:, 0])

    # Grid.

    x = np.expand_dims(np.linspace(start=xmin, stop=xmax, num=numx), axis=1)

    # Plot sample.

    if scatter_feature is not None:
        scatter_feature(pdf=pdf, x=x)
        scatter_color = scatter_feature.coloring()
    else:
        scatter_color = None

    if ax is None:
        _, ax = plt.subplots()

    if rand is None:
        rand = np.random

    ax.scatter(
        sample[:, 0],
        rand.random(size=len(sample)),
        c=scatter_color.to_rgba(scatter_feature.calculate_values(sample))
        if scatter_color is not None
        else "w",
        s=scatter_size,
        edgecolors="k",
    )

    ax.set_yticks([])
    ax.set_ylim(0.0, 1.0)

    if scatter_color is not None:
        plt.colorbar(mappable=scatter_color, shrink=colorbar_shrink)

    # Plot feature contour.

    if contour_feature is not None:
        contour_feature(pdf=pdf, x=x)

        zeros = feature_roots(
            feature=contour_feature,
            xmin=xmin,
            xmax=xmax,
            fsolve_x0_size=fsolve_x0_size,
        )

        ax.vlines(
            x=zeros,
            ymin=0.0,
            ymax=1.0,
            linestyles=["--"] * len(zeros),
            colors=["k"] * len(zeros),
        )

    return ax


def plot_sample_2d(
    pdf: PDF,
    scatter_feature: PDFFeature = None,
    contour_features: List[PDFSignedFeature] = None,
    contour_colors: List[str] = None,
    contour_linestyles: List[str] = None,
    xmin: float = None,
    xmax: float = None,
    numx: int = 101,
    ymin: float = None,
    ymax: float = None,
    numy: int = 101,
    sample: np.ndarray = None,
    size: int = None,
    rand: np.random.RandomState = None,
    scatter_size: float = None,
    contour_linewidth: float = None,
    contour_opacity: float = 1.0,
    fill_contour: bool = False,
):
    """
    Plots a random sample from a given bivariate pdf along with zero-level
    contours of one or more pdf features.

    :param pdf: Bivariate density to plot.
    :param scatter_feature: Feature to color scatter dots.
    :param contour_features: pdf features to extract zero-level contours.
    :param contour_colors: Colors for the contours.
    :param contour_linestyles: Line styles for the contours.
    :param xmin: Minimum value in the x axis.
    :param xmax: Maximum value in the x axis.
    :param numx: Number of points in the x axis.
    :param ymin: Minimum value in the y axis.
    :param ymax: Maximum value in the y axis.
    :param numy: Number of points in the y axis.
    :param sample: Sample to plot.
    :param size: Number of random observations to draw from the pdf.
    :param rand: Random generator.
    :param scatter_size: Size of observation markers.
    :param contour_linewidth: Width of feature contours.
    :param contour_opacity: A number between 0 (transparent) and 1 (opaque).
    :param fill_contour: Whether to fill the contour. Only used when a single
     contour is present.
    :return: The generated axes.
    """

    if pdf.dimension != 2:
        raise ValueError("Density is not 2-dimensional")

    if sample is None:
        if size is None:
            raise ValueError(
                "Either a sample or a sample size must be provided"
            )

        sample = pdf.sample(size=size, rand=rand)

    # Bounds.

    if xmin is None:
        xmin = np.amin(sample[:, 0])

    if xmax is None:
        xmax = np.amax(sample[:, 0])

    if ymin is None:
        ymin = np.amin(sample[:, 1])

    if ymax is None:
        ymax = np.amax(sample[:, 1])

    # Grid.

    xy, (x, y) = rectangular_grid((xmin, xmax, numx), (ymin, ymax, numy))

    # Plot sample.

    if scatter_feature is not None:
        scatter_feature(pdf=pdf, x=xy)
        scatter_color = scatter_feature.coloring()
    else:
        scatter_color = None

    _, ax = plt.subplots()

    ax.scatter(
        x=sample[:, 0],
        y=sample[:, 1],
        color=scatter_color.to_rgba(scatter_feature.calculate_values(sample))
        if scatter_color is not None
        else "w",
        s=scatter_size,
        edgecolors="k",
    )

    # Plot feature contours.

    if contour_features is None:
        contour_features = []

    if contour_colors is None:
        contour_colors = []

    if contour_linestyles is None:
        contour_linestyles = ["solid"] * len(contour_features)

    if not (
        len(contour_features) == len(contour_colors) == len(contour_linestyles)
    ):
        raise ValueError("Different sizes for contour parameters")

    for contour_feature, color, line_style in zip(
        contour_features, contour_colors, contour_linestyles
    ):
        contour_feature(pdf=pdf, x=xy)

        contour_config = {
            "colors": color,
            "linestyles": line_style,
            "alpha": contour_opacity,
        }

        if len(contour_features) == 1 and fill_contour:
            contour = ax.contourf
            levels = sorted([0.0, contour_features[0].reference_sign * np.Inf])
        else:
            contour = ax.contour
            levels = [0.0]
            contour_config["linewidths"] = contour_linewidth

        contour_config["levels"] = levels

        contour(x, y, contour_feature.values.reshape(x.shape), **contour_config)

    if scatter_color is not None:
        plt.colorbar(mappable=scatter_color, location="left")

    ax.set_xlim(left=xmin, right=xmax)
    ax.set_ylim(bottom=ymin, top=ymax)

    return ax


def plot_sample_3d(
    pdf: PDF,
    scatter_feature: PDFFeature = None,
    contour_features: List[PDFSignedFeature] = None,
    contour_color_maps: List[str] = None,
    contour_transparencies: List[float] = None,
    xmin: float = None,
    xmax: float = None,
    numx: int = 101,
    ymin: float = None,
    ymax: float = None,
    numy: int = 101,
    zmin: float = None,
    zmax: float = None,
    numz: int = 101,
    sample: np.ndarray = None,
    size: int = None,
    rand: np.random.RandomState = None,
    scatter_size: float = 2.0,
    contour_linewidth: float = 0.05,
    contour_line_color: str = "k",
    colorbar_shrink: float = 0.5,
    elevation: float = None,
    azimuth: float = None,
):
    """
    Plots a random sample from a given trivariate pdf along with zero-level
    contours of one or more pdf features.

    :param pdf: Trivariate density to plot.
    :param scatter_feature: Feature to color scatter dots.
    :param contour_features: Features to determine contours.
    :param contour_color_maps: Color maps for contours.
    :param contour_transparencies: Transparency levels of contours.
    :param xmin: Minimum value in the x axis.
    :param xmax: Maximum value in the x axis.
    :param numx: Number of points in the x axis.
    :param ymin: Minimum value in the y axis.
    :param ymax: Maximum value in the y axis.
    :param numy: Number of points in the y axis.
    :param zmin: Minimum value in the z axis.
    :param zmax: Maximum value in the z axis.
    :param numz: Number of points in the z axis.
    :param sample: Sample to plot.
    :param size: Number of random observations to draw from the pdf.
    :param rand: Random generator.
    :param scatter_size: Size of observation markers.
    :param contour_linewidth: Width of feature contours.
    :param contour_line_color: Color of feature contours.
    :param colorbar_shrink: Shrink factor for color bar.
    :param elevation: The elevation angle in the vertical plane in degrees.
    :param azimuth: The azimuth angle in the horizontal plane in degrees.
    :return: The generated axes.
    """

    if pdf.dimension != 3:
        raise ValueError("Density is not 3-dimensional")

    if sample is None:
        if size is None:
            raise ValueError(
                "Either a sample or a sample size must be provided"
            )

        sample = pdf.sample(size=size, rand=rand)

    # Bounds.

    if xmin is None:
        xmin = np.amin(sample[:, 0])

    if xmax is None:
        xmax = np.amax(sample[:, 0])

    if ymin is None:
        ymin = np.amin(sample[:, 1])

    if ymax is None:
        ymax = np.amax(sample[:, 1])

    if zmin is None:
        zmin = np.amin(sample[:, 2])

    if zmax is None:
        zmax = np.amax(sample[:, 2])

    # Grid.

    xyz, (x, y, z) = rectangular_grid(
        (xmin, xmax, numx), (ymin, ymax, numy), (zmin, zmax, numz)
    )

    # Plot sample.

    fig = plt.figure()
    ax = fig.gca(projection="3d")
    ax.view_init(elev=elevation, azim=azimuth)

    if scatter_feature is not None:
        scatter_feature(pdf=pdf, x=xyz)
        coloring = scatter_feature.coloring()
    else:
        coloring = None

    ax.scatter(
        sample[:, 0],
        sample[:, 1],
        sample[:, 2],
        c=coloring.to_rgba(scatter_feature.calculate_values(sample))
        if coloring is not None
        else None,
        s=scatter_size,
    )

    # Plot 3D contour.

    if contour_features is None:
        contour_features = []

    if contour_color_maps is None:
        contour_color_maps = []

    if contour_transparencies is None:
        contour_transparencies = []

    if not (
        len(contour_features)
        == len(contour_color_maps)
        == len(contour_transparencies)
    ):
        raise ValueError("Different sizes for contour parameters")

    for contour_feature, color_map, transparency in zip(
        contour_features, contour_color_maps, contour_transparencies
    ):
        contour_feature(pdf=pdf, x=xyz)

        values = contour_feature.values.reshape(x.shape)

        vertices, faces, _, _ = marching_cubes(
            values,
            0.0,
            spacing=(
                (xmax - xmin) / numx,
                (ymax - ymin) / numy,
                (zmax - zmin) / numz,
            ),
            allow_degenerate=False,
        )

        vertices += np.array([xmin, ymin, zmin])

        ax.plot_trisurf(
            vertices[:, 0],
            vertices[:, 1],
            faces,
            vertices[:, 2],
            cmap=color_map,
            linewidth=contour_linewidth,
            edgecolors=contour_line_color,
            alpha=transparency,
        )

    if coloring is not None:
        plt.colorbar(mappable=coloring, location="left", shrink=colorbar_shrink)

    return ax
