"""
Density surface plotting with additional features.
"""
from typing import Final

import numpy as np
from matplotlib import pyplot as plt

from kdfeatures.density import PDF
from kdfeatures.plot import PDFFeature, PDFSignedFeature, PDFUnsignedHeight
from kdfeatures.plot.feature import feature_roots
from kdfeatures.util.grid import rectangular_grid


def plot_surface_1d(
    pdf: PDF,
    xmin: float,
    xmax: float,
    xnum: int,
    area_feature: PDFFeature = PDFUnsignedHeight(),
    fsolve_x0_size: int = 21,
):
    """
    Plots the graph of a univariate pdf.

    :param pdf: Density to be plotted.
    :param xmin: Minimum value of the domain.
    :param xmax: Maximum value of the domain.
    :param xnum: Number of points for interpolation.
    :param area_feature: Density feature to plot as colors under the curve.
    :param fsolve_x0_size: Number of initial guesses to fin all roots of the
     area feature function.
    """
    if pdf.dimension != 1:
        raise ValueError("pdf must be 1-dimensional")

    # Height.

    x = np.linspace(start=xmin, stop=xmax, num=xnum)
    x_matrix = np.expand_dims(x, axis=1)
    y = pdf(x_matrix)

    # Feature.

    area_feature(pdf=pdf, x=x_matrix)
    area_values = np.squeeze(area_feature.values)
    area_color = area_feature.coloring()

    # Area.

    _, ax = plt.subplots()

    for _x, _y, c in zip(x, y, area_color.to_rgba(area_values)):
        ax.vlines(x=_x, ymin=0.0, ymax=_y, colors=c)

    plt.colorbar(mappable=area_color, location="right")

    # Curve.

    ax.plot(x, y, "k-")
    ax.set_xlim(left=xmin, right=xmax)
    ax.set_ylim(bottom=0.0)

    # Zeros.

    if isinstance(area_feature, PDFSignedFeature):
        zeros = feature_roots(
            feature=area_feature,
            xmin=xmin,
            xmax=xmax,
            fsolve_x0_size=fsolve_x0_size,
        )

        pdf_zeros = np.squeeze(pdf(np.expand_dims(zeros, axis=1)))
        ax.plot(zeros, pdf_zeros, "kx")

        ax.vlines(
            x=zeros,
            ymin=0.0,
            ymax=pdf_zeros,
            linestyles=["--"] * len(zeros),
            colors=["k"] * len(zeros),
        )


def plot_surface_2d(
    pdf: PDF,
    xmin: float,
    xmax: float,
    numx: int,
    ymin: float,
    ymax: float,
    numy: int,
    xstride: int = 1,
    ystride: int = 1,
    area_feature: PDFFeature = PDFUnsignedHeight(),
    colorbar_shrink: float = 0.7,
    elevation: float = None,
    azimuth: float = None,
):
    """
    Plots a bidimensional pdf as a colored surface.

    :param pdf: Density function to plot.
    :param xmin: Minimum value in the x axis.
    :param xmax: Maximum value in the x axis.
    :param numx: Number of points in the x axis.
    :param ymin: Minimum value in the y axis.
    :param ymax: Maximum value in the y axis.
    :param xstride: Sampling rate for graphing in the x axis.
    :param ystride: Sampling rate for graphing in the y axis.
    :param numy: Number of points in the y axis.
    :param area_feature: Density feature to plot as colors over the
     surface area.
    :param colorbar_shrink: Shrink factor for color bar.
    :param elevation: The elevation angle in the vertical plane in degrees.
    :param azimuth: The azimuth angle in the horizontal plane in degrees.
    """
    if pdf.dimension != 2:
        raise ValueError("pdf must be 2-dimensional")

    # Rectangular grid.

    xy, (x, y) = rectangular_grid((xmin, xmax, numx), (ymin, ymax, numy))

    # Height.

    height = PDFUnsignedHeight()
    height(pdf=pdf, x=xy)
    height_values = height.values.reshape(x.shape)

    max_height: Final = np.amax(height_values)

    # Area color.

    if isinstance(area_feature, PDFUnsignedHeight):
        area = height
        area_values = height_values
    else:
        area_feature(pdf=pdf, x=xy)
        area = area_feature
        area_values = area.values.reshape(x.shape)

    area_color = area.coloring()

    # Surface plot.

    fig = plt.figure()

    ax = fig.gca(projection="3d")

    ax.grid(False)
    ax.view_init(elev=elevation, azim=azimuth)

    ax.plot_surface(
        x,
        y,
        height_values,
        facecolors=area_color.to_rgba(area_values),
        rstride=xstride,
        cstride=ystride,
        linewidth=0,
        antialiased=False,
    )

    # Contour plot for signed area features.

    if isinstance(area_feature, PDFSignedFeature):
        ax.contour(
            x,
            y,
            area_values,
            zdir="z",
            levels=[0.0],
            colors="m",
            linestyles="solid",
            offset=max_height,
        )

    ax.set_zlim(0.0, max_height)

    plt.colorbar(mappable=area_color, location="left", shrink=colorbar_shrink)

    return ax
