"""
Tools for subsetting large arrays.
"""
from itertools import chain

import numpy as np

from kdfeatures.util.matrix import cartesian_differences


def split_range(num_elements: int, num_batches: int):
    """
    Splits a range 0:n in approximately equal size ranges.

    :param num_elements: Number elements in the range.
    :param num_batches: Approximate number of batches.
    :return: A generator of approximately equal size ranges.
    """
    whole_batch_size, remainder = divmod(num_elements, num_batches)
    return chain(
        (
            range(whole_batch_size * k, whole_batch_size * (k + 1))
            for k in range(num_batches)
        ),
        [range(num_elements - remainder, num_elements)]
        if remainder > 0
        else iter(()),
    )


def double_sum_batches(sample: np.ndarray, num_batches: int):
    """
    Returns batches of the cartesian product of a sample with itself.

    :param sample: Sample to generate batches from.
    :param num_batches: Number of batches.
    :return: A generator of batches for the double sum cartesian differences of
     a random sample.
    """
    if num_batches is None:
        return [cartesian_differences(sample, sample)]
    else:
        ranges = split_range(num_elements=len(sample), num_batches=num_batches)
        return (cartesian_differences(sample[r], sample) for r in ranges)
