"""
Operations with partial derivative multi-indices.
"""
from collections import Counter, OrderedDict
from math import comb
from typing import Final, List, Tuple

import numpy as np


class HermiteMinimalPolynomial:
    def __init__(self, z: np.ndarray, v: np.ndarray, order: int):
        # Argument checking.

        if order < 0:
            raise ValueError("Wrong polynomial order")

        if z.shape[1] != v.shape[0]:
            raise ValueError("Incompatible dimensions between z and v")

        # Internals.

        self._z: Final = z
        self._v: Final = v
        self._k: Final = order

        self._values: Final = OrderedDict()

        # Initialization.

        if order == 0:
            self[(0,) * self.dimension] = np.ones(shape=(len(self.z), 1))
        elif order == 1:
            ms: Final = MultiIndices(d=self.dimension, r=1).unique_multi_indices

            for m, z in zip(ms, self.z.transpose()):
                self[m] = z.reshape((-1, 1))

    @property
    def dimension(self) -> int:
        return self.z.shape[1]

    @property
    def order(self) -> int:
        return self._k

    @property
    def z(self) -> np.ndarray:
        return self._z

    @property
    def v(self) -> np.ndarray:
        return self._v

    def __len__(self) -> int:
        return len(self._values)

    def __getitem__(self, m: Tuple[int]) -> np.ndarray:
        return self._values[m]

    def __setitem__(self, m: Tuple[int], value: np.ndarray):
        if m in self._values:
            raise ValueError("Value already present")

        self._values[m] = value

    def enlarged_vector(self) -> np.ndarray:
        indices: Final = MultiIndices(d=self.dimension, r=self.order)

        return np.column_stack(
            [
                list(self._values.values())[
                    indices.expanded_to_compressed_derivative_index(i)
                ]
                for i in range(self.dimension**self.order)
            ]
        )


def hermite_minimal_polynomial(
    x: np.ndarray, v: np.ndarray, order: int
) -> HermiteMinimalPolynomial:

    z: Final = np.dot(x, v)

    if order == 0 or order == 1:
        return HermiteMinimalPolynomial(z=z, v=v, order=order)

    h_k_minus_1 = HermiteMinimalPolynomial(z=z, v=v, order=0)
    h_k = HermiteMinimalPolynomial(z=z, v=v, order=1)

    for k in range(2, order + 1):
        h_aux = h_k
        h_k = build_hermite_k_plus_1(h_k_minus_1=h_k_minus_1, h_k=h_k)
        h_k_minus_1 = h_aux

    return h_k


def build_hermite_k_plus_1(
    h_k_minus_1: HermiteMinimalPolynomial, h_k: HermiteMinimalPolynomial
) -> HermiteMinimalPolynomial:

    if h_k_minus_1.order != h_k.order - 1:
        raise ValueError("Incompatible polynomial orders")

    h_k_plus_1: Final = HermiteMinimalPolynomial(
        z=h_k.z, v=h_k.v, order=h_k.order + 1
    )

    ms: Final = MultiIndices(d=h_k.dimension, r=h_k.order).unique_multi_indices

    for j in range(0, h_k.dimension):
        n = num_elements(d=h_k.dimension - j, r=h_k.order)

        for m in ms[-n:]:
            assert m[0:j] == (0,) * j

            m_plus_1 = add_one_at_position(m=m, pos=j)
            value = recursive_hermite(
                h_k_minus_1=h_k_minus_1, h_k=h_k, m=m, j=j
            )

            h_k_plus_1[m_plus_1] = value

    assert len(h_k_plus_1) == num_elements(d=h_k.dimension, r=h_k.order + 1)

    return h_k_plus_1


def num_elements(d: int, r: int) -> int:
    return comb(r + d - 1, r)


def recursive_hermite(
    h_k_minus_1: HermiteMinimalPolynomial,
    h_k: HermiteMinimalPolynomial,
    m: Tuple[int],
    j: int,
) -> np.ndarray:
    acc = np.zeros(shape=(len(h_k.z), 1))

    for i in range(j, h_k.dimension):
        if m[i] > 0:
            acc += (
                h_k.v[j, i]
                * m[i]
                * h_k_minus_1[substract_one_at_position(m=m, pos=i)]
            )

    return h_k.z[:, j].reshape((-1, 1)) * h_k[m] - acc


def add_one_at_position(m: Tuple[int], pos: int) -> Tuple:
    return m[0:pos] + ((m[pos] + 1),) + m[(pos + 1) :]


def substract_one_at_position(m: Tuple[int], pos: int) -> Tuple:
    return m[0:pos] + ((m[pos] - 1),) + m[(pos + 1) :]


class MultiIndices:
    def __init__(self, d: int, r: int):
        tuples: Final = multi_indices(d=d, r=r)
        tuple_orders: Final = dict()

        k = 0

        for t in tuples:
            if t not in tuple_orders:
                tuple_orders[t] = k
                k += 1

        self._tuple_orders: Final = [tuple_orders[t] for t in tuples]
        self._unique_multi_indices: Final = list(OrderedDict.fromkeys(tuples))

    @property
    def unique_multi_indices(self) -> List[Tuple]:
        return self._unique_multi_indices

    def expanded_to_compressed_derivative_index(self, i: int) -> int:
        return self._tuple_orders[i]


def multi_indices(d: int, r: int) -> List[Tuple]:
    return [tuple(row) for row in _p_matrix_no_duplicates(d=d, r=r)]


def _p_matrix_no_duplicates(d: int, r: int) -> np.ndarray:
    p: Final = _p_matrix(d, r)
    return np.apply_along_axis(
        func1d=lambda x: _multi_index_no_duplicates(x=x, d=d), axis=1, arr=p
    )


def _p_matrix(d: int, r: int) -> np.ndarray:
    q: Final = _q_matrix(d=d, r=r)
    return (np.floor(q[:, :-1]) - d * np.floor(q[:, 1:])).astype(int)


def _q_matrix(d: int, r: int) -> np.ndarray:
    q: Final = np.zeros(shape=(d**r, r + 1))

    for i in range(q.shape[0]):
        for j in range(q.shape[1]):
            q[i, j] = i / (d**j)

    return q


def _multi_index_no_duplicates(x: np.ndarray, d: int):
    counts: Final = Counter(x)
    return np.array([counts[i] if i in counts else 0 for i in range(d)])
