"""
Basic matrix functionality.
"""
from typing import Final

import numpy as np
from scipy.linalg import eigh, eigvals


def is_matrix(array: np.ndarray) -> bool:
    """
    Checks whether an input array is a matrix.

    :param array: Array to check.
    :return: true if and only if the array has two dimensions.
    """
    return len(array.shape) == 2


def is_square(matrix: np.ndarray) -> bool:
    """
    Checks whether an input tensor is a square matrix.

    :param matrix: Input tensor.
    :return: true if and only if the input tensor is a square matrix.
    """
    return is_matrix(matrix) and matrix.shape[0] == matrix.shape[1]


def is_symmetric(matrix: np.ndarray) -> bool:
    """
    Checks whether a matrix is symmetric.

    :param matrix: Matrix to be tested.
    :return: true if and only if the input matrix is symmetric.
    """
    return np.allclose(a=matrix, b=matrix.T)


def is_positive_definite(matrix: np.ndarray) -> bool:
    """
    Checks whether a square matrix is positive definite, that is, if all its
    eigenvalues are greater than zero.

    :param matrix: Matrix to be tested.
    :return: true if and only if the input matrix is positive definite.
    """
    return np.all(eigvals(matrix) > 0)


def is_negative_definite(matrix: np.ndarray) -> bool:
    """
    Checks whether a square matrix is negative definite, that is, if all its
    eigenvalues are smaller than zero.

    :param matrix: Matrix to be tested.
    :return: true if and only if the input matrix is negative definite.
    """
    return largest_eigenvalue(matrix) < 0


def is_inner_product(matrix: np.ndarray) -> bool:
    """
    Checks whether an input matrix is an inner product matrix, that is,
    square, symmetric and positive definite.

    :param matrix: Matrix to be tested.
    :return: true if and only if the input matrix is an inner product matrix.
    """
    return (
        is_square(matrix)
        and is_symmetric(matrix)
        and is_positive_definite(matrix)
    )


def is_negative_inner_product(matrix: np.ndarray) -> bool:
    """
    Checks whether an input matrix is a negative inner product matrix, that is,
    square, symmetric and negative definite.

    :param matrix: Matrix to be tested.
    :return: true if and only if the input matrix is a negative inner product
     matrix.
    """
    return (
        is_square(matrix)
        and is_symmetric(matrix)
        and is_negative_definite(matrix)
    )


def largest_eigenvalue(matrix: np.ndarray) -> float:
    """
    Calculates the largest eigenvalue.

    :param matrix: Matrix whose largest eigenvalue needs to be calculated.
    :return: The largest eigenvalue in the input matrix.
    """

    if not is_square(matrix):
        raise ValueError("Matrix is not square")

    highest = matrix.shape[0] - 1

    return float(eigh(matrix, eigvals_only=True, eigvals=(highest, highest))[0])


def smallest_eigenvalue(matrix: np.ndarray) -> float:
    """
    Calculates the smallest eigenvalue.

    :param matrix: Matrix whose smallest eigenvalue needs to be calculated.
    :return: The smallest eigenvalue in the input matrix.
    """

    if not is_square(matrix):
        raise ValueError("Matrix is not square")

    return float(eigh(matrix, eigvals_only=True, eigvals=(0, 0))[0])


def remove_nan_rows(matrix: np.ndarray) -> np.ndarray:
    """
    Removes rows where any of the elements is not a number.

    :param matrix: Matrix to clean.
    :return: The input matrix with the nan-rows removed.
    """
    return matrix[~np.isnan(matrix).any(axis=1), :]


def cartesian_differences(x: np.ndarray, y: np.ndarray) -> np.ndarray:
    """
    Given two matrices of row-wise vectors X_i and Y_j, builds a new matrix
    of elements X_i - Y_j, where the i index changes more slowly that the j.

    :param x: First matrix operand.
    :param y: Second matrix operand.
    :return: A row-wise vector matrix of differences.
    """
    if not is_matrix(x):
        raise ValueError("x is not a matrix")

    if not is_matrix(y):
        raise ValueError("y is not a matrix")

    if x.shape[1] != y.shape[1]:
        raise ValueError("x and y do not have the same number of columns")

    return np.repeat(x, repeats=[len(y)] * len(x), axis=0) - np.tile(
        y, reps=(len(x), 1)
    )


def from_log_cholesky(v: np.ndarray) -> np.ndarray:
    """
    Builds a positive-definite matrix from a log-Cholesky parameterization.

    :param v: Log-Cholesky parameterization of the matrix.
    :return: The positive-definite matrix resulting from the input log-Cholesky
     parameterization.
    """
    n: Final = int((-1 + np.sqrt(1 + 8 * len(v))) / 2)
    m: Final = np.zeros((n, n))

    m[np.triu_indices_from(m)] = v

    for i in range(m.shape[0]):
        m[i, i] = np.exp(m[i, i])

    return np.dot(m.transpose(), m)


def to_log_cholesky(m: np.ndarray) -> np.ndarray:
    """
    Obtains the log-Cholesky parameterization of a positive-definite matrix.

    :param m: Matrix whose log-Cholesky parameterization is required.
    :return: The log-Cholesky parameterization of the input matrix.
    """
    if not is_positive_definite(m):
        raise ValueError("Matrix is not positive-definite")

    upper_tri: Final = np.linalg.cholesky(m).transpose()

    for i in range(upper_tri.shape[0]):
        if upper_tri[i, i] < 0:
            upper_tri[i, :] *= -1

        upper_tri[i, i] = np.log(upper_tri[i, i])

    return upper_tri[np.triu_indices_from(upper_tri)]
