"""
Test on the Gaussian density.
"""
from typing import Final

import numpy as np
import pytest
from kdfeatures.density.gaussian import (
    Gaussian,
    gaussian_hessian,
    gaussian_jacobian,
    gaussian_pdf,
)

from kdfeatures.util.matrix import is_negative_inner_product


@pytest.mark.parametrize(
    "x",
    [
        np.array([[0.0]]),
        np.array([[0.0, 1.0], [1.0, 0.0]]),
        np.array([[0.0, 1.0, 2.0], [1.0, 2.0, 0.0], [2.0, 0.0, 1.0]]),
    ],
)
def test_gaussian_pdf(x: np.ndarray):
    """
    Checks that the Gaussian pdf works properly for different input tensor
    shapes.

    :param x: Input tensor for testing purposes.
    """
    pdf: Final = gaussian_pdf(x)
    assert pdf.shape == (len(x),)


@pytest.mark.parametrize(
    "x",
    [
        np.array([[0.0]]),
        np.array([[0.0, 1.0], [1.0, 0.0]]),
        np.array([[0.0, 1.0, 2.0], [1.0, 2.0, 0.0], [2.0, 0.0, 1.0]]),
    ],
)
def test_gaussian_jacobian(x: np.ndarray):
    """
    Checks that the Gaussian Jacobian works properly for different input tensor
    shapes.

    :param x: Input tensor for testing purposes.
    """
    jac: Final = gaussian_jacobian(x)
    assert jac.shape == (len(x), x.shape[1])


@pytest.mark.parametrize(
    "x",
    [
        np.array([[0.0]]),
        np.array([[0.0, 1.0], [1.0, 0.0]]),
        np.array([[0.0, 1.0, 2.0], [1.0, 2.0, 0.0], [2.0, 0.0, 1.0]]),
    ],
)
def test_gaussian_hessian(x: np.ndarray):
    """
    Checks that the Gaussian Hessian works properly for different input tensor
    shapes.

    :param x: Input tensor for testing purposes.
    """
    hess: Final = gaussian_hessian(x)
    assert hess.shape == (len(x), x.shape[1] ** 2)


@pytest.mark.parametrize(
    "gaussian",
    [
        Gaussian(1, mean=np.array([1.0]), covariance=np.array([[1.0]])),
        Gaussian(
            2,
            mean=np.array([0.0, 1.0]),
            covariance=np.array([[1.0, 0.5], [0.5, 1.0]]),
        ),
        Gaussian(
            3,
            mean=np.array([0.0, 1.0, -1.0]),
            covariance=np.array(
                [[1.0, 0.5, 0.3], [0.5, 1.0, 0.7], [0.3, 0.7, 1.0]]
            ),
        ),
    ],
)
def test_gaussian(gaussian: Gaussian):
    mean = np.array([gaussian.mean])

    # Check Jacobian at the mean is identically zero.

    jac = gaussian.jacobian(mean)

    for derivative in jac.flatten():
        assert derivative == pytest.approx(0.0, abs=1e-6)

    # Check Hessian at the mean is a negative inner product.

    hess = gaussian.hessian(mean).reshape(
        (gaussian.dimension, gaussian.dimension)
    )

    assert is_negative_inner_product(hess)

    # Check largest eigenvalue.

    assert float(gaussian.largest_hessian_eigenvalue(mean)) < 0
