"""
Check risk measure calculations.
"""
from typing import Final

import numpy as np
import pytest
from scipy.stats import norm

from conftest import random_state
from kdfeatures.density import univariate_gaussian
from kdfeatures.util.quant import tail_value_at_risk


@pytest.mark.parametrize("p", np.linspace(0, 1, 11)[1:-1])
def test_tail_value_at_risk(p: float, sample_size: int = 1_000_000):
    """
    Check TVaR calculations.

    :param p: Confidence for the TVaR.
    :param sample_size: Sample size for the approximated TVaR.
    """
    # Get a standard normal sample.

    sample: Final = univariate_gaussian().sample(
        size=sample_size, rand=random_state
    )

    # Calculate exact TVaR.

    true_tvar: Final = norm.pdf(norm.ppf(q=p)) / (1 - p)

    # Calculate approximate TVaR.

    approx_tvar: Final = tail_value_at_risk(x=sample, p=p)

    # Compare.

    assert true_tvar == pytest.approx(approx_tvar, rel=1e-2)
